#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
"""
ACUTE-Eval Analyzer.

Given a run_id, one can print out and save the results.

If you specify the pairings filepath, you can visualize
the choices that turkers made about specific conversations.
"""
import os
import pandas as pd
import numpy as np

from typing import Dict, Any, List, Optional
from scipy.stats import binom_test

from parlai.mturk.core.mturk_data_handler import MTurkDataHandler
from parlai import __file__ as parlai_filepath
from parlai.core.params import ParlaiParser
import json
from IPython.core.display import HTML


# throw away turkers below this threshold
AGREEMENT_THRESHOLD = 0.8
# do we count ties as agreements?
AGREEMENT_TIES_OKAY = False


def setup_args():
    """
    Setup appropriate args.
    """
    parser = ParlaiParser(True, False)
    parser.add_argument(
        '-id', '--run-id', type=str, default=None, help='run id to analyze'
    )
    parser.add_argument(
        '--is-sandbox',
        type='bool',
        default=True,
        help='whether the run is a sandbox run or not',
    )
    parser.add_argument(
        '--outdir', type=str, default=None, help='where to save the results'
    )
    parser.add_argument(
        '--pairings-filepath',
        type=str,
        default=None,
        help='path to the acute analysis pairs for the corresponding run id',
    )
    parser.add_argument(
        '--rounding-digit',
        type=int,
        default=2,
        help='number of digits for rounding displayed table',
    )
    parser.add_argument(
        '--max-matchups-html',
        type=int,
        default=10,
        help='max number of matchups to display per model pair in html',
    )
    parser.add_argument(
        '--min-dialogue-length',
        type=int,
        default=-1,
        help='the minimum number of turns for both dialogues in a matchup to be counted as valid for analysis',
    )
    parser.add_argument(
        '--annotate-convo',
        type='bool',
        default=False,
        help='whether to include a checkbox column for annotating the conversation pairs',
    )
    return parser


def _print_progress(msg: str):
    """
    Format a msg to print to stdout well.

    :param msg:
        message to print
    """
    print(f"\n{'-' * 60}\n {msg} \n {'-' * 60}")


class AcuteAnalyzer(object):
    """
    Analyzer.

    Given a run_id, we can do lots of fun things!
    """

    def __init__(self, opt: Dict, remove_failed: bool = True):
        """
        Initialize the analyzer.

        Builds up the dataframe

        :param opt:
            opt dict

        :param remove_failed:
            Whether to remove ratings from turkers who failed onboarding
        """
        self.ROOT_DIR = os.path.join(opt['datapath'], 'acute_eval')
        self.run_id = opt['run_id']
        self.is_sandbox = opt['is_sandbox']
        self.outdir = opt['outdir']
        self.pairings_filepath = opt['pairings_filepath']
        if not self.pairings_filepath:
            self.pairings_filepath = ''
        if not self.outdir:
            self.outdir = os.path.join(self.ROOT_DIR, f'{self.run_id}-results')
        if not os.path.exists(self.ROOT_DIR):
            os.mkdir(self.ROOT_DIR)
        if not os.path.exists(self.outdir):
            os.mkdir(self.outdir)
        self.db_path = os.path.join(
            os.path.dirname(parlai_filepath),
            'mturk',
            'run_data',
            f"pmt_{'sb' if opt['is_sandbox'] else ''}data.db",
        )
        self.dataframe = self._extract_to_dataframe()
        self.min_dialogue_length = opt.get('min_dialogue_length', -1)
        self.max_matchups_html = opt.get('max_matchups_html', 10)
        self.annotate_convo = opt.get('annotate_convo', False)
        if remove_failed:
            self._remove_failed_onboarding()
        self._extract_model_names()
        self._load_pairings_file()

    def _extract_response_data(
        self,
        data: Dict[str, Any],
        task_data: Dict[str, Any],
        hit: Dict[str, Any],
        response_data: Dict[str, Any],
    ) -> Dict[str, Any]:
        """
        Extract response data from task data.

        :param data:
            full data from a given turker
        :param task_data:
            data from one "task" completion (i.e. one dialogue comparison)
        :param hit:
            hit data
        :param response_data:
            turker's corresponding response data corresponding to the task

        :return response:
            Turker's response data from the task
        """
        response: Dict[str, Any] = {
            'run_id': self.run_id,
            'worker': data['worker_id'],
            'time_taken': hit['task_end'] - hit['task_start'],
            'question': data['task_data'][0]['task_specs']['question'],
            'conversation_id': hit['conversation_id'],
        }
        onboarding = task_data['task_specs'].get('is_onboarding', False)
        choice = response_data['speakerChoice']
        if onboarding:
            response['correct'] = choice == task_data['pairing_dict']['correct_answer']
        else:
            response['correct'] = -1

        speakers_to_eval = sorted(task_data["pairing_dict"]["speakers_to_eval"])
        response.update(
            {
                'winner': choice,
                'loser': speakers_to_eval[1 - (speakers_to_eval.index(choice))],
                'eval_choice_0': speakers_to_eval[0],
                'eval_choice_1': speakers_to_eval[1],
                'reason': response_data['textReason'],
                'is_onboarding': onboarding,
                'matchup': f"{'__vs__'.join(speakers_to_eval)}",
                'pairing_id': task_data['pair_id'],
                'dialogue_lengths': {
                    task_data['task_specs']['model_left']['name']: len(
                        task_data['task_specs']['model_left']['dialogue']
                    ),
                    task_data['task_specs']['model_right']['name']: len(
                        task_data['task_specs']['model_right']['dialogue']
                    ),
                },
                'speaker_model_mapping': [
                    task_data['task_specs']['model_left']['name'],
                    task_data['task_specs']['model_right']['name'],
                ],
            }
        )
        return response

    def _get_hit_data(
        self, hit: Dict[str, Any], logger: MTurkDataHandler
    ) -> Optional[Dict[str, Any]]:
        """
        Return data for a given hit.

        If the HIT is corrupt for whatever reason, we return None

        :param hit:
            HIT information dict
        :param logger:
            Data handler

        :return data:
            Optional dict with the hit data
        """
        try:
            full_data: Dict[str, Any] = logger.get_full_conversation_data(
                self.run_id, hit['conversation_id'], self.is_sandbox
            )
        except FileNotFoundError:
            print(
                f"WARNING: Data for run_id `{self.run_id}` not found for "
                f"conversation id {hit['conversation_id']}"
            )
            return None

        data: Dict[str, Any] = next(iter(full_data['worker_data'].values()))
        if not (
            'task_data' in data['response'] and len(data['response']['task_data']) > 0
        ):
            # worker abandoned task, drop their annotations
            return None
        elif len(data['response']['task_data']) != len(data['task_data']):
            raise ValueError('Saved task data does not match response task data')

        return data

    def _extract_to_dataframe(self) -> pd.DataFrame:
        """
        Extract the data from the run to a pandas dataframe.
        """
        logger = MTurkDataHandler(file_name=self.db_path)
        hits = logger.get_pairings_for_run(self.run_id)

        dataframe: List[Dict[str, Any]] = []
        for hit in hits:
            if hit['conversation_id'] is None:
                continue
            data = self._get_hit_data(hit, logger)
            if data is None:
                continue
            for r_idx, task_data in enumerate(data['task_data']):
                response_data = data['response']['task_data'][r_idx]
                if response_data is None or not response_data:
                    continue
                response = self._extract_response_data(
                    data, task_data, hit, response_data
                )
                dataframe.append(response)
        return pd.DataFrame(dataframe)

    def _remove_failed_onboarding(self):
        """
        Remove workers who failed onboarding.
        """
        df = self.dataframe

        all_workers_failing_onboarding = df.loc[
            df['is_onboarding'] & (df['correct'] == False), 'worker'  # noqa: E712
        ].values

        workers_failing_onboarding = sorted(
            np.unique(all_workers_failing_onboarding).tolist()
        )

        self.dataframe = df[
            ~df["worker"].isin(workers_failing_onboarding) & ~df["is_onboarding"]
        ]

    def _load_pairings_file(self):
        """
        Load the pairings file, if provided.

        Allows for visualization of the conversations turkers rated.
        """
        df = self.dataframe
        if not self.pairings_filepath or not os.path.exists(self.pairings_filepath):
            return
        pairings = []
        with open(self.pairings_filepath, 'r') as f:
            for line in f:
                pair = json.loads(line)
                model1, model2 = pair['speakers_to_eval']
                pair[model1] = pair['dialogue_dicts'][0]
                pair[model2] = pair['dialogue_dicts'][1]
                del pair['dialogue_dicts']
                pairings.append(pair)
        pairs_to_eval = [pairings[i] for i in df.pairing_id.values.tolist()]
        # Build dialogue_ids => dialogue mappings

        winner_dialogues = []
        loser_dialogues = []
        for i, (_, row) in enumerate(df.iterrows()):
            winner = row['winner']
            loser = row['loser']
            winner_dialogues.append(pairs_to_eval[i][winner])
            loser_dialogues.append(pairs_to_eval[i][loser])
        df.loc[:, 'pairs_to_eval'] = pd.Series(pairs_to_eval, index=df.index)
        df.loc[:, 'winner_dialogue'] = pd.Series(winner_dialogues, index=df.index)
        df.loc[:, 'loser_dialogue'] = pd.Series(loser_dialogues, index=df.index)
        self.dataframe = df

    def _extract_model_names(self):
        """
        Extract the model nicknames from the dataframe.
        """
        df = self.dataframe
        df = df[df['run_id'] == self.run_id]
        matchups = list(df.matchup.unique())
        models = set()
        combos = set()
        for matchup in matchups:
            model1, model2 = matchup.split('__vs__')
            models.add(model1)
            models.add(model2)
            combos.add(tuple(sorted((model1, model2))))
        self.models = list(models)
        self.models.sort()
        self.combos = list(combos)
        self.combos.sort()

    ###############################
    # Data manipulation Functions #
    ###############################
    def get_reasons(self) -> List[str]:
        """
        Return dataframe reasons.
        """
        return self.dataframe['reason'].values.tolist()

    def get_max_hits_per_worker(self) -> List[int]:
        """
        Get max number of hits per worker.
        """
        return self.dataframe.groupby('worker')['run_id'].count().max()

    def filter_by_dialogue_length(self, is_debug=False):
        """
        Filter out matchup with one of the conversation shorter than
        self.min_dialogue_length This applies to calculating sorted_win_frac_df and
        signficance_df, but not html visualizations of conversations.

        :param is_debug: if True, print logs indicating the number of pairings filtered out due to short conversation.
            is_debug bool
        """
        df = pd.DataFrame()
        filter_list = {}
        for _, row in self.dataframe.iterrows():
            keep_row = True
            for model_name, dialogue_length in row['dialogue_lengths'].items():
                if keep_row and dialogue_length < self.min_dialogue_length:
                    keep_row = False
                    filter_list[model_name] = filter_list.get(model_name, 0) + 1
            if keep_row:
                df = df.append(row, ignore_index=True)
        if is_debug:
            for model_name in filter_list:
                print(
                    f"For {self.run_id}: filter out {filter_list[model_name]} matchups due to {model_name} with dialogue length shorter than {self.min_dialogue_length}"
                )
        return df

    def get_wins_per_model_matchup(self) -> pd.DataFrame:
        """
        Return the wins for each model by matchup.
        """
        df_filtered = self.filter_by_dialogue_length(True)
        self.matchup_total_df = (
            df_filtered.groupby(['eval_choice_0', 'eval_choice_1'])['run_id']
            .count()
            .to_frame('matchup_total')
        )
        self.win_total_df = (
            df_filtered.groupby(['eval_choice_0', 'eval_choice_1', 'winner', 'loser'])[
                'loser'
            ]
            .count()
            .to_frame('win_total')
            .reset_index()
            .set_index(['eval_choice_0', 'eval_choice_1'])
        )
        return self.win_total_df

    def get_win_fractions(self) -> pd.DataFrame:
        """
        Return the joined matchup + win totals, get win fractions.

        Sorted according to win percentage
        """
        if not hasattr(self, 'win_total_df'):
            self.get_wins_per_model_matchup()

        self.win_fraction_df = self.matchup_total_df.join(self.win_total_df).assign(
            win_frac=lambda df: df['win_total'] / df['matchup_total']
        )

        pivoted_df = self.win_fraction_df.pivot(
            index="loser", columns="winner", values="win_frac"
        )
        self.models_by_win_frac = (
            self.win_fraction_df.groupby("winner")["win_frac"]
            .mean()
            .sort_values()
            .index.values.tolist()
        )
        self.sorted_win_frac_df = pivoted_df.reindex(
            index=self.models_by_win_frac, columns=self.models_by_win_frac
        )
        return self.sorted_win_frac_df

    def get_num_hits_per_matchup(self):
        """
        Return the number of hits per matchup.
        """
        matchup_total_1_df = self.matchup_total_df.reset_index()
        matchup_total_2_df = matchup_total_1_df.rename(
            columns={'eval_choice_0': 'eval_choice_1', 'eval_choice_1': 'eval_choice_0'}
        )
        self.num_hits_per_matchup_df = (
            pd.concat([matchup_total_1_df, matchup_total_2_df], axis=0)
            .pivot(
                index='eval_choice_0', columns='eval_choice_1', values='matchup_total'
            )
            .reindex(index=self.models_by_win_frac, columns=self.models_by_win_frac)
        )
        return self.num_hits_per_matchup_df

    ##################
    # Rendering HTML #
    ##################
    def render_conversations_per_matchups(self):
        """
        Render conversations with and without reasons included.
        """
        matchups = list(self.dataframe.matchup.unique())

        def _render_row(matchup: List[str], row: pd.Series, row_id: int) -> str:
            dialogues = {'winner_dialogue': '', 'loser_dialogue': ''}
            for d_key in dialogues:
                result = []
                for _, turn in enumerate(row[d_key]['dialogue']):
                    speakername = turn['id']
                    text = turn['text']
                    is_bot = (
                        (speakername != 'human_evaluator')
                        and (speakername != 'other_speaker')
                        and speakername in matchup
                    )
                    align = 'right' if is_bot else 'left'
                    color = "white" if is_bot else "black"
                    bgcolor = '#2391f7' if is_bot else '#e1e1e7'

                    result.append(
                        (
                            '<div style="overflow: auto; padding: 1ex 0;">'
                            f'<div style="clear: both; float: {align}; color: {color}; background-color: {bgcolor}; padding: 0.5em 1em; border-radius: 1em; max-width: 80%">'
                            f'<p style="margin: 0">{speakername}: {text}</p>'
                            '</div>'
                            '</div>'
                        )
                    )
                dialogues[d_key] = (
                    '<div style="background-color: white; margin: 0em; padding: 0.5em; '
                    'font-family: sans-serif; font-size: 9pt; width: 99%;">'
                    + ''.join(result)
                    + '</div>'
                )

            if row['winner'] == row['speaker_model_mapping'][0]:
                speakers_footnote = "(Speaker_1[winner] = {}, Speaker_2 = {})".format(
                    row['speaker_model_mapping'][0], row['speaker_model_mapping'][1]
                )
            else:
                speakers_footnote = "(Speaker_1 = {}, Speaker_2[winner] = {})".format(
                    row['speaker_model_mapping'][0], row['speaker_model_mapping'][1]
                )

            checkbox_row = (
                '<td>'
                '<div><input type= "checkbox" id= "cherry" name= "cherry">'
                '<label for="cherry">Cherry</label>'
                '</div>'
                '<div><input type= "checkbox" id= "lemon" name= "lemon">'
                '<label for= "lemon">Lemon</label>'
                '</div>'
                '<div><input type= "checkbox" id= "neutral" name= "neutral">'
                '<label for= "neutral">Neutral</label>'
                '</div>'
                '</td>'
            )
            dialogue_row = f"<td>{dialogues['winner_dialogue']}</td><td>{dialogues['loser_dialogue']}</td>"
            reason_row = f"<td>{row['reason']}\n{speakers_footnote}</td>"
            if self.annotate_convo:
                return f"<tr><td>Pair {str(row_id)}</td>{checkbox_row}{dialogue_row}{reason_row}</tr>"
            else:
                return f"<tr><td>Pair {str(row_id)}</td>{dialogue_row}{reason_row}</tr>"

        def _render_html(table: pd.DataFrame) -> HTML:
            result = '\
                <div id="toc_container">\
                <p class="toc_title">Model Pairs</p>\
                <ul class="toc_list">'
            for matchup in matchups:
                eval_question = table.loc[table['matchup'] == matchup, 'question'].iloc[
                    0
                ]
                result += f"<li><a href='#{matchup}''>{matchup + '__on__' +eval_question}</a></li>"
            result += '</ul></div>'
            for matchup in matchups:
                length = min(
                    self.max_matchups_html, len(table[table['matchup'] == matchup])
                )
                eval_question = table.loc[table['matchup'] == matchup, 'question'].iloc[
                    0
                ]
                matchup_table = table[table['matchup'] == matchup][:length]
                table_rows = [
                    _render_row(matchup.split('__vs__'), row, idx)
                    for idx, (_, row) in enumerate(matchup_table.iterrows())
                ]
                if self.annotate_convo:
                    table_body = f"<table border=1 frame=void rules=rows cellpadding='20'><tr><th>Pair Id</th><th>Comments</th><th>Winner Conversation</th><th>Loser Conversation</th><th>Reason</th></tr>{''.join(table_rows)}</table>"
                else:
                    table_body = f"<table border=1 frame=void rules=rows cellpadding='20'><tr><th>Pair Id</th><th>Winner Conversation</th><th>Loser Conversation</th><th>Reason</th></tr>{''.join(table_rows)}</table>"
                result += f"<h2 id='{matchup}'><li><a href='#toc_container'>{matchup + '__on__' +eval_question}</a></li></h2><body>{table_body}</body>"
            return HTML(result)

        table = self.dataframe

        self.rendered_without_reasons = _render_html(table)
        table = table[table['reason'] != '']
        self.rendered_with_reasons = _render_html(table)

    #############################
    # Functionality from Notebook
    #############################
    def get_matchup_totals_with_signficance(self) -> pd.DataFrame:
        """
        Return dataframe with matchup win totals + significance.
        """

        def _signf_level(p):
            if p < 0.001:
                return "***", "p<.001"
            elif p < 0.01:
                return "**", "p<.01"
            elif p < 0.05:
                return "*", "p<.05"
            else:
                return "", "p>.05"

        output = []
        df_filtered = self.filter_by_dialogue_length()
        for _, run_annotations in df_filtered.groupby('run_id'):
            question = list(run_annotations.question)[0]
            for matchup, annotations in run_annotations.groupby('matchup'):
                model1, model2 = matchup.split('__vs__')
                wincount1 = np.sum(annotations['winner'] == model1)
                wincount2 = np.sum(annotations['winner'] == model2)
                winrate1 = np.mean(annotations['winner'] == model1)
                winrate2 = np.mean(annotations['winner'] == model2)
                p = binom_test([wincount1, wincount2])

                stars, plevel = _signf_level(p)

                agreements = []
                for _, pairing_annotations in annotations.groupby('pairing_id'):
                    pair_wincount1 = np.sum(pairing_annotations['winner'] == model1)
                    pair_wincount2 = np.sum(pairing_annotations['winner'] == model2)
                    if pair_wincount1 < 2 and pair_wincount2 < 2:
                        if pair_wincount1 == 1 and pair_wincount2 == 1:
                            agreements.append(0)
                    else:
                        majority_wincount = max(pair_wincount1, pair_wincount2)
                        num_pair_annotations = pair_wincount1 + pair_wincount2
                        pair_agreement = majority_wincount / num_pair_annotations
                        agreements.append(pair_agreement)
                total_agreement = np.mean(agreements)

                output.append(
                    {
                        'question': question,
                        'matchup': matchup,
                        'model1': model1,
                        'model2': model2,
                        'numwins1': wincount1,
                        'numwins2': wincount2,
                        'winrate1': winrate1,
                        'winrate2': winrate2,
                        'p': p,
                        'stars': stars,
                        'sigf': plevel,
                        'agree': total_agreement,
                    }
                )
        df_output = pd.DataFrame(output)
        # order the columns how we want
        self.signficance_df = df_output[
            [
                'question',
                'matchup',
                'model1',
                'numwins1',
                'winrate1',
                'model2',
                'numwins2',
                'winrate2',
                'sigf',
                'stars',
                'p',
                'agree',
            ]
        ]
        return self.signficance_df

    def save_results(self, path: str = None):
        """
        Save results to a certain path.

        This function saves the dataframe results to a csv.

        In addition, if the pairings filepath is specified, this function will also
        save rendered conversations to an html file for viewing.
        """
        if not hasattr(self, 'signficance_df'):
            self.get_matchup_totals_with_signficance()
        if path is None:
            path = self.outdir

        def _path(filename):
            return os.path.join(path, f"{self.run_id}.{filename}")

        # Save significance file
        with open(_path('significance.csv'), 'w') as f:
            f.write(self.signficance_df.to_csv(index=False))
        _print_progress(
            f"To visualize signficance result, try cat {_path('significance.csv')} | column -t -s, | less -S"
        )
        # Save Win Grid
        with open(_path('grid.csv'), 'w') as f:
            f.write(self.get_win_fractions().to_csv(index=True))
        _print_progress(
            f"To visualize grid result, try cat {_path('grid.csv')} | sed 's/,/ ,/g' | column -t -s, | less -S"
        )

        # Render conversations if valid pairings filepath provided
        if os.path.exists(self.pairings_filepath):
            if not hasattr(self, 'rendered_with_reasons'):
                self.render_conversations_per_matchups()
            with open(_path('reason.html'), 'w') as f:
                f.write(self.rendered_with_reasons.data)
            _print_progress(
                f"To visualize only conversations with reasons provided, open {_path('reason.html')} in a browser"
            )
            with open(_path('all.html'), 'w') as f:
                f.write(self.rendered_without_reasons.data)
            _print_progress(
                f"To visualize all conversations, open {_path('all.html')} in a browser"
            )


if __name__ == "__main__":
    parser = setup_args()
    args = parser.parse_args()
    analyzer = AcuteAnalyzer(args)
    results = pd.DataFrame(analyzer.get_win_fractions())
    analyzer.save_results()
    _print_progress('Here is the grid of results:')
    rounding_digit = args.get('rounding_digit', 2)
    print(results.round(rounding_digit).to_string())
    result = pd.DataFrame(analyzer.get_matchup_totals_with_signficance())
    result = result.drop(columns=['agree'])
    _print_progress('Here is each matchup with signficance measures:')
    print(result.round(rounding_digit).to_string())



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import json
import math
import os
import torch
import torch.nn.functional as F
from collections import defaultdict, Counter
from nltk import ngrams

from parlai.core.torch_generator_agent import PPLMetric
from parlai.agents.image_seq2seq.image_seq2seq import ImageSeq2seqAgent
from parlai.agents.transformer.transformer import TransformerGeneratorAgent
from parlai.core.metrics import AverageMetric, SumMetric, GlobalAverageMetric
from parlai.utils.misc import round_sigfigs
from parlai.utils.io import PathManager


def div(x, y):
    if y == 0:
        return x
    else:
        return x / y


class NGramIterator:
    """
    N-Gram iterator for a list.
    """

    def __init__(self, lst, n):
        self.lst = lst
        self.n = n
        self.max = len(lst) - n

    def __iter__(self):
        self.counter = -1
        return self

    def __next__(self):
        self.counter += 1
        if self.counter > self.max:
            raise StopIteration
        return tuple(self.lst[self.counter : self.counter + self.n])


class RewardUnlikelihoodAgentTrait(object):
    """
    Abstract Trait.

    Applies unlikelihood loss based on the presence of negative rewards in the task.
    """

    @classmethod
    def add_cmdline_args(cls, argparser):
        grp = super(RewardUnlikelihoodAgentTrait, cls).add_cmdline_args(argparser)
        grp.add_argument('--alpha', default=1.0, type=float)

    def batchify(self, obs_batch, **kwargs):
        batch = super().batchify(obs_batch, **kwargs)
        rewards = torch.FloatTensor(
            [float(o.get('reward', 0)) for o in batch.observations]
        ).to(batch.text_vec.device)
        batch['rewards'] = rewards
        return batch

    def _dummy_batch(self, batchsize, maxlen):
        batch = super()._dummy_batch(batchsize, maxlen)
        batch['rewards'] = torch.ones(batchsize, dtype=torch.long).cuda()
        return batch

    def compute_loss(self, batch, return_output=False):
        if batch.label_vec is None:
            raise ValueError('Cannot compute loss without a label.')

        model_output = self.model(*self._model_input(batch), ys=batch.label_vec)
        scores, preds, *_ = model_output  # scores is bsz x time x vocab

        scores = F.log_softmax(scores, dim=-1)
        scores_view = scores.view(-1, scores.size(-1))
        targets = batch.label_vec
        targets_view = targets.view(-1)

        notnull = targets.ne(self.NULL_IDX)
        if self.is_training:
            # note it's >= because convai2 and other teachers all provide a 0 reward
            mle_notnull = notnull & (batch.rewards >= 0).unsqueeze(1).expand_as(notnull)
        else:
            mle_notnull = notnull

        mle_loss = (
            F.nll_loss(
                scores_view, targets_view, ignore_index=self.NULL_IDX, reduction='none'
            ).view_as(mle_notnull)
            * mle_notnull.float()
        ).sum()

        # limit loss to only the positive rewards
        mle_target_tokens = mle_notnull.long().sum()
        correct = ((targets == preds) * mle_notnull).sum()
        self.global_metrics.add('token_acc', AverageMetric(correct, mle_target_tokens))
        self.global_metrics.add('nll_loss', AverageMetric(mle_loss, mle_target_tokens))
        self.global_metrics.add('ppl', PPLMetric(mle_loss, mle_target_tokens))
        if mle_target_tokens > 0:
            mle_loss /= mle_target_tokens  # average loss per token

        if not self.is_training:
            if return_output:
                return (mle_loss, model_output)
            else:
                return mle_loss

        # and now we want the unlikelihood loss on the negative examples
        ul_notnull = notnull & (batch.rewards < 0).unsqueeze(1).expand_as(notnull)
        ul_target_tokens = ul_notnull.long().sum()
        range_ = torch.arange(targets_view.size(0)).to(batch.label_vec.device)
        ul_scores = scores_view[range_, targets_view]
        clamp_min = 1e-6 if self.opt['fp16'] else 1e-20
        ul_loss = (
            -torch.log(torch.clamp(1.0 - ul_scores.exp(), min=clamp_min)).view_as(
                ul_notnull
            )
            * ul_notnull.float()
        ).sum()
        self.global_metrics.add('ul_loss', AverageMetric(ul_loss, ul_target_tokens))
        if ul_target_tokens > 0:
            ul_loss /= ul_target_tokens

        loss = mle_loss + self.opt['alpha'] * ul_loss

        if return_output:
            return (loss, model_output)
        else:
            return loss


class RepetitionUnlikelihoodAgentTrait(object):
    """
    Abstract Trait.

    Applies unliikelihood loss to repetition some proportion of train steps by
    generating, marking repeats, and calculating loss accordingly.
    """

    def __init__(self, opt, shared=None):
        super().__init__(opt, shared)
        self.pred_logsoftmax = torch.nn.LogSoftmax(dim=2)

    @classmethod
    def add_cmdline_args(cls, argparser):
        print(super())
        grp = super().add_cmdline_args(argparser)
        grp.add_argument('--seq-ul-ratio', default=0.5, type=float)
        grp.add_argument('--seq-ul-n', default=4, type=int)
        grp.add_argument('--mask-n', default=100, type=int)
        grp.add_argument('--ctxt-beta', default=0.5, type=float)
        grp.add_argument('--crep-pen', default='crep', type=str)

    def _init_cuda_buffer(self, batchsize, maxlen, force=False):
        pass

    def _count_n_grams(self, token_lst, n):
        n_grams = defaultdict(int)
        for n_gram in NGramIterator(token_lst, n):
            n_grams[n_gram] += 1
        return n_grams

    def compute_loss(self, batch, return_output=False):
        if self.is_training and (torch.rand(1).item() >= self.opt['seq_ul_ratio']):
            total_loss, model_output = super().compute_loss(batch, return_output=True)
            # No sequence level unlikelihood
            if return_output:
                return total_loss, model_output
            else:
                return total_loss

        # Generate
        clamp_min = 1e-6 if self.opt['fp16'] else 1e-20
        maxlen = self.label_truncate or 256
        with torch.no_grad():
            beam_pred_scores, _ = self._generate(batch, self.beam_size, maxlen)

        # forward pass to create graph for beam search case
        generations = [g[1:] for (g, s) in beam_pred_scores]
        pred_toks = torch.nn.utils.rnn.pad_sequence(generations, batch_first=True)
        model_output = self.model(*self._model_input(batch), ys=pred_toks)
        logits, preds, _ = model_output

        # construct mask marking repeats
        n = self.opt['seq_ul_n']  # label n-grams
        crep_mask = torch.zeros_like(pred_toks).type_as(logits)
        lrep_mask = torch.zeros_like(pred_toks).type_as(logits)

        for i, gen in enumerate(generations):
            gen_i = gen.tolist()

            # Collect context ngrams
            context_i = batch.text_vec[i].tolist()
            context_n_grams = self._count_n_grams(context_i, n)

            seen_n_grams = defaultdict(int)

            # penalize if there is a context repeat
            for j, n_gram in enumerate(NGramIterator(gen_i, n)):
                if context_n_grams[n_gram] > 0:
                    crep_mask[i, j : j + n] = 1

            # penalize if there is a label repeat
            for j, n_gram in enumerate(NGramIterator(gen_i, n)):
                if seen_n_grams[n_gram] > 0:
                    lrep_mask[i, j : j + n] = 1
                seen_n_grams[n_gram] += 1

        # Compute unlikelihood loss
        lprobs = self.pred_logsoftmax(logits)
        pred_lprobs = lprobs.view(-1, lprobs.size(2)).gather(1, pred_toks.view(-1, 1))
        one_minus_probs = torch.clamp((1.0 - pred_lprobs.exp()), min=clamp_min).view(
            pred_toks.size(0), pred_toks.size(1)
        )

        mask = ((1 - self.opt['ctxt_beta']) * lrep_mask) + (
            self.opt['ctxt_beta'] * crep_mask
        )

        ul_loss = -(torch.log(one_minus_probs)) * mask
        total_loss = div(ul_loss.sum(), mask.sum())
        self.record_local_metric(
            'ul_loss', AverageMetric.many(ul_loss.sum(dim=-1), mask.sum(dim=-1))
        )

        if not self.is_training:
            # in eval mode, we want metrics (e.g. PPL) provided by tga's compute_loss
            _, _ = super().compute_loss(batch, return_output=True)

        if return_output:
            return total_loss, model_output
        return total_loss

    def _add_generation_metrics(self, batch, preds):
        self._ngram_metrics(batch, preds)

    def _ngram_metrics(self, batch, preds):
        text_vecs_cpu = batch.text_vec.cpu()
        lrep, crep = 0, 0
        total_pred_ngs = 0
        n = self.opt['seq_ul_n']
        for i, pred in enumerate(preds):
            pred_token_list = pred.tolist()
            if self.END_IDX in pred_token_list:
                pred_token_list = pred_token_list[
                    : pred_token_list.index(self.END_IDX)
                ]  # remove possible padding
            if self.START_IDX in pred_token_list:
                pred_token_list = pred_token_list[
                    pred_token_list.index(self.START_IDX) :
                ]
            pred_ngs = [ng for ng in ngrams(pred_token_list, n)]
            pred_counter = Counter(pred_ngs)
            total_pred_ngs += len(pred_ngs)
            lrep += len(pred_ngs) - len(pred_counter)

            text_token_list = text_vecs_cpu[i].tolist()
            if self.NULL_IDX in text_token_list:
                text_token_list = text_token_list[
                    : text_token_list.index(self.NULL_IDX)
                ]  # remove possible padding
            context_counter = Counter([ng for ng in ngrams(text_token_list, n)])

            for ng in pred_counter:
                if ng in context_counter:
                    crep += pred_counter[ng]

        self.global_metrics.add(
            'lrep_%dgrams' % n, GlobalAverageMetric(lrep, total_pred_ngs)
        )
        self.global_metrics.add(
            'crep_%dgrams' % n, GlobalAverageMetric(crep, total_pred_ngs)
        )


class _VocabUnlikelihoodTrait(object):
    def reset_metrics(self):
        super().reset_metrics()
        self.metrics['high_freq'] = 0
        self.metrics['gold_high'] = 0

    def _kldiv(self, p_counter, q_counter) -> float:
        ptotal = sum(p_counter.values())
        qtotal = sum(q_counter.values())
        kldiv = 0.0
        for word, _ in p_counter.items():
            prob_p = p_counter[word] / ptotal
            prob_q = q_counter[word] / qtotal
            kldiv += prob_p * math.log(1e-20 + (prob_q / prob_p))
        return -kldiv

    def _jsdiv(self, dist1: Counter, dist2: Counter) -> float:
        half = dist1 + dist2
        return 0.5 * self._kldiv(dist1, half) + 0.5 * self._kldiv(dist2, half)

    def report(self):
        report = super().report()
        report['kldiv_humgen'] = self._kldiv(
            self.running_human, self.running_generation
        )
        report['kldiv_genhum'] = self._kldiv(
            self.running_generation, self.running_human
        )
        report['jsdiv'] = self._jsdiv(self.running_human, self.running_generation)
        return report


class SequenceVocabUnlikelihoodAgentTrait(_VocabUnlikelihoodTrait):
    """
    Abstract Trait.

    Applies unlikelihood loss to vocabulary distributiion by generating, calculating
    proportion of tokens per vocabulary frequency bin, and computing loss accordingly
    """

    def __init__(self, opt, shared=None):
        super().__init__(opt, shared)
        self.NUM_STEPS = opt['queue_size']
        if shared is None:
            self._reset_running_histories()
            self._last_was_training = True
            self.truebins = {}
            counts_file = self.opt['counts_file']
            if counts_file is None:
                counts_file = os.path.join(
                    os.path.dirname(self.opt['model_file']), 'counts.txt'
                )
                if not PathManager.exists(counts_file):
                    raise RuntimeError(
                        'Please give a --counts-file to use vocab unlikelihood'
                    )
            with PathManager.open(counts_file) as f:
                for line in f:
                    record = json.loads(line)
                    self.truebins[record['word_id']] = record['bin']

    def reset(self):
        super().reset()
        self._reset_running_histories()

    def _reset_running_histories(self):
        self.generation_history = []
        self.running_generation = Counter()
        self.human_history = []
        self.running_human = Counter()

    @classmethod
    def add_cmdline_args(cls, argparser):
        print(super())
        grp = super().add_cmdline_args(argparser)
        grp.add_argument('--alpha', default=1.0, type=float)
        grp.add_argument('--queue-size', default=32, type=int)
        grp.add_argument(
            '--weighting', choices={'uniform', 'logdiff', 'kldiv'}, default='uniform'
        )
        grp.add_argument('--threshold', type=float, default=1e-3)
        grp.add_argument('--counts-file', type=str, default=None)

    def _init_cuda_buffer(self, *args, **kwargs):
        pass

    def reset_metrics(self):
        super().reset_metrics()
        self.metrics['num_penalize'] = 0
        self.metrics['steps'] = 0
        self.metrics['hum_toks'] = 0
        self.metrics['gen_toks'] = 0
        self.metrics['ul_weights'] = 0
        # better clean our statistics so not to leak test statistics into train

    def _get_bins(self, counts: Counter):
        c = Counter()
        for k, v in counts.items():
            c.update({self.truebins.get(k, 'never'): v})
        t = sum(c.values())
        return {k: round_sigfigs(v / t, 4) for k, v in c.items()}

    def _l2dist(self, bins):
        return (
            (bins.get('frequent', 0) - 0.4) ** 2
            + (bins.get('medium', 0) - 0.3) ** 2
            + (bins.get('rare', 0) - 0.2) ** 2
            + (bins.get('veryrare', 0) - 0.1) ** 2
            + (bins.get('never', 0) - 0.0) ** 2
        )

    def report(self):
        r = super().report()
        if self.running_generation and self.running_human:
            for k, v in self._get_bins(self.running_human).items():
                r[f'humdist_{k}'] = v
            gendist = self._get_bins(self.running_generation)
            for k, v in gendist.items():
                r[f'gendist_{k}'] = v
            r['dist_l2'] = self._l2dist(gendist)

        return r

    def compute_loss(self, batch, return_output=False):
        if self._last_was_training is not self.is_training:
            self._reset_running_histories()
            self._last_was_training = self.is_training

        nll_loss, model_output = super().compute_loss(batch, True)
        scores, preds, *_ = model_output  # scores is bsz x time x vocab
        targets = batch.label_vec
        notnull = targets != self.NULL_IDX

        with torch.no_grad():
            beam_pred_scores, _ = self._generate(
                batch, self.beam_size, self.opt['label_truncate']
            )

            # forward pass to create graph for beam search case
            generations = [g for (g, s) in beam_pred_scores]
            gentoks = torch.nn.utils.rnn.pad_sequence(
                generations, batch_first=True, padding_value=self.NULL_IDX
            )
            # strip the BOS tokens
            gentoks = gentoks[:, 1:]

        # find everything we oversampled
        gen_mask = gentoks != self.NULL_IDX
        self.generation_history.append(Counter(gentoks[gen_mask].view(-1).tolist()))
        self.human_history.append(Counter(targets[notnull].view(-1).tolist()))
        self.running_generation += self.generation_history[-1]
        self.running_human += self.human_history[-1]

        if len(self.generation_history) > self.NUM_STEPS:
            if not self.is_training:
                # we want a running history of word usage
                self.running_generation -= self.generation_history.pop(0)
                self.running_human -= self.human_history.pop(0)
        else:
            if return_output:
                return nll_loss, model_output
            else:
                return nll_loss

        gen_sum = sum(self.running_generation.values())
        hum_sum = sum(self.running_human.values())

        # what did we oversample?
        if self.opt['weighting'] == 'logdiff':
            to_penalize = {
                w: (v / gen_sum) - (self.running_human.get(w, 0) / hum_sum)
                for w, v in self.running_generation.items()
            }
            to_penalize = {
                w: v for w, v in to_penalize.items() if v >= self.opt['threshold']
            }
            to_penalize = {w: math.log(v / 0.001) for w, v in to_penalize.items()}
        elif self.opt['weighting'] == 'uniform':
            to_penalize = {
                w: (v / gen_sum) - (self.running_human.get(w, 0) / hum_sum)
                for w, v in self.running_generation.items()
            }
            to_penalize = {
                w: 1 for w, v in to_penalize.items() if v >= self.opt['threshold']
            }
        elif self.opt['weighting'] == 'kldiv':
            to_penalize = {
                w: (
                    self.running_generation[w] / gen_sum,
                    self.running_human[w] / hum_sum,
                )
                for w, v in self.running_human.items()
                if w in self.running_generation
            }
            to_penalize = {
                w: (p_gen, p_hum)
                for w, (p_gen, p_hum) in to_penalize.items()
                if p_gen > p_hum
            }
            to_penalize = {
                w: p_gen * (math.log(p_gen) - math.log(p_hum))
                for w, (p_gen, p_hum) in to_penalize.items()
            }
            to_penalize = {
                k: v for k, v in to_penalize.items() if v > self.opt['threshold']
            }
        else:
            raise ValueError

        self.global_metrics.add('num_penalize', SumMetric(len(to_penalize)))

        ul_weights = torch.zeros(gen_mask.shape)
        ul_mask = torch.zeros_like(gen_mask)
        for wordid, weight in to_penalize.items():
            ul_mask = ul_mask | (gentoks == wordid)
            ul_weights[gentoks == wordid] = weight
        ul_weights = ul_weights.to(gen_mask.device)
        self.global_metrics.add('ul_weights', AverageMetric(ul_weights[ul_mask].mean()))

        # and whack it
        model_output = self.model(*self._model_input(batch), ys=gentoks)
        scores, *_ = model_output
        downweight = gentoks[ul_mask]

        almost_scores = F.log_softmax(scores[ul_mask], dim=-1)
        ul_scores = almost_scores[torch.arange(len(downweight)), downweight]

        clamp_min = 1e-6 if self.opt['fp16'] else 1e-20

        ul_loss = (
            -(torch.log(torch.clamp(1 - ul_scores.exp(), min=clamp_min)))
            * ul_weights[ul_mask]
        ).sum()
        num_ul = ul_mask.sum()

        self.global_metrics.add('ul_loss', AverageMetric(ul_loss, num_ul))
        self.global_metrics.add('ul_num_tokens', SumMetric(num_ul))

        ul_loss = div(ul_loss, num_ul)

        if len(self.generation_history) < self.NUM_STEPS:
            loss = nll_loss
        else:
            loss = nll_loss + self.opt['alpha'] * ul_loss

        if return_output:
            return (loss, model_output)
        else:
            return loss


class TransformerSequenceVocabUnlikelihoodAgent(
    SequenceVocabUnlikelihoodAgentTrait, TransformerGeneratorAgent
):
    """
    Example usage:

    -t convai2 -m parlai_internal.projects.dontsaythat.agents:TransformerSequenceVocabUnlikelihoodAgent
    """

    pass


class TransformerUnlikelihoodAgent(
    RewardUnlikelihoodAgentTrait, TransformerGeneratorAgent
):
    """
    Example usage:

    -t convai2 -m parlai_internal.projects.dontsaythat.agents:TransformerUnlikelihoodAgent
    """

    pass


class RepetitionUnlikelihoodAgent(RepetitionUnlikelihoodAgentTrait, ImageSeq2seqAgent):
    """
    Example usage:

    -t convai2 -m parlai_internal.projects.dontsaythat.agents:RepetitionUnlikelihoodAgent
    """

    pass



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from parlai.utils.testing import AutoTeacherTest  # noqa: F401
import unittest


class TestDefaultTeacher(unittest.TestCase, AutoTeacherTest):
    task = 'convai2'


class TestNormalizedTeacher(unittest.TestCase, AutoTeacherTest):
    task = 'convai2:normalized'


class TestBothTeacher(unittest.TestCase, AutoTeacherTest):
    task = 'convai2:both'


class TestNoneTeacher(unittest.TestCase, AutoTeacherTest):
    task = 'convai2:none'


class TestSelfRevisedTeacher(unittest.TestCase, AutoTeacherTest):
    task = 'convai2:self_revised'



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from parlai.core.teachers import ParlAIDialogTeacher
from .build import build, get_fpath

import copy
import os
import random


def _path(opt):
    # Build the data if it doesn't exist.
    build(opt)
    # for now train, valid and test will be identical, will change with more data.
    dt = opt['datatype'].split(':')[0]
    fpath = get_fpath(opt)

    return os.path.join(
        opt['datapath'],
        'light_dialogue_wild',
        fpath,
        opt['light_label_type'] + '_' + dt + '.txt',
    )


class DefaultTeacher(ParlAIDialogTeacher):
    @staticmethod
    def add_cmdline_args(argparser):
        agent = argparser.add_argument_group('LIGHT Dialogue options')
        agent.add_argument(
            '--light_use_repeat',
            type=str,
            default='none',
            choices=['self_last', 'partner_last', 'none', 'both_last'],
        )
        agent.add_argument('--light_use_taskname', type='bool', default=True)
        agent.add_argument('--light_use_setting', type='bool', default=True)
        agent.add_argument('--light_use_unseen_test', type='bool', default=False)
        agent.add_argument('--light_use_person_names', type='bool', default=True)
        agent.add_argument(
            '--light_use_persona',
            type=str,
            default='self',
            choices=['partner', 'self', 'all', 'none'],
        )
        agent.add_argument(
            '--light_use_emote',
            type=str,
            default='all',
            choices=['partner', 'self', 'all', 'none'],
        )
        agent.add_argument(
            '--light_use_speech',
            type=str,
            default='all',
            choices=['partner', 'self', 'all', 'none'],
        )
        agent.add_argument(
            '--light_use_action',
            type=str,
            default='all',
            choices=['partner', 'self', 'all', 'none'],
        )
        agent.add_argument('--light_use_affordances', type='bool', default=True)
        agent.add_argument(
            '--light_use_current_self_output',
            type=str,
            default="all",
            choices=['none', 'all', 'all_filtered', 'all_filtered_remove'],
        )
        agent.add_argument(
            '--light_label_type',
            type=str,
            default='speech',
            choices=['speech', 'action', 'emote'],
            help='type of target in light dialogues',
        )
        agent.add_argument('--light_use_cands', type=int, default=20)
        agent.add_argument('--light_use_clip_cands', type=int, default=10000)
        agent.add_argument('--light_use_speech_prefix', type='bool', default=True)
        agent.add_argument('--light_use_score_cutoff', type=int, default=1)
        agent.add_argument('--light_use_max_score_cutoff', type=int, default=-1)
        agent.add_argument(
            '--light_use_hard_score_cutoff',
            type='bool',
            default=False,
            help='Specify True to **only** include examples with the specified '
            'score cutoff. E.g., if `--light-use-score-cutoff 3`, will only '
            'supply dialogues with scores of 3.',
        )
        agent.add_argument(
            '--light-model-name',
            type=str,
            default=None,
            help='if specified, the model from which chats should be saved '
            'to specify multiple, delimit with + symbol',
        )
        agent.add_argument(
            '--light-use-continue-type',
            type=str,
            default='all',
            choices=['all', 'continue', 'exit'],
            help='only use dialogues that had a specific continue outcome, default use all',
        )
        agent.add_argument(
            '--light-use-date-cutoff',
            type=str,
            default=None,
            help=(
                'If specified, only include hobbot conversations that were collected '
                'before this date. Format YYYY-MM-DD.\n'
                'First wild experiments ended 2020-05-26.\n'
                'First model group ended 2020-06-23.\n'
            ),
        )
        agent.add_argument(
            '--light-use-person-names-prefix',
            type='bool',
            default=False,
            help="If specified, prefix text with character name",
        )
        agent.add_argument('--light_percent_train_exs', type=float, default=1.0)

    def __init__(self, opt, shared=None):
        opt = copy.deepcopy(opt)
        opt['parlaidialogteacher_datafile'] = _path(opt)
        if 'light_use_speech_prefix' not in opt:
            opt['light_use_speech_prefix'] = True
        self.pct_train_exs = opt['light_percent_train_exs']
        super().__init__(opt, shared)

    def _setup_data(self, path):
        """
        Overriding to limit num train exs.
        """
        super()._setup_data(path)
        if self.training and self.pct_train_exs <= 1.0:
            random.seed(42)
            self.episodes = random.sample(
                self.episodes, int(self.num_episodes() * self.pct_train_exs)
            )
            self.num_exs = sum(len(e) for e in self.episodes)


class SimpleTeacher(DefaultTeacher):
    @staticmethod
    def add_cmdline_args(argparser):
        agent = argparser.add_argument_group('LIGHT Dialogue options')
        agent.add_argument(
            '--light_use_repeat',
            type=str,
            default='none',
            choices=['self_last', 'partner_last', 'none', 'both_last'],
        )
        agent.add_argument('--light_use_taskname', type='bool', default=True)
        agent.add_argument('--light_use_setting', type='bool', default=True)
        agent.add_argument('--light_use_unseen_test', type='bool', default=False)
        agent.add_argument('--light_use_person_names', type='bool', default=True)
        agent.add_argument(
            '--light_use_persona',
            type=str,
            default='self',
            choices=['partner', 'self', 'all', 'none'],
        )
        agent.add_argument(
            '--light_use_emote',
            type=str,
            default='none',
            choices=['partner', 'self', 'all', 'none'],
        )
        agent.add_argument(
            '--light_use_speech',
            type=str,
            default='partner',
            choices=['partner', 'self', 'all', 'none'],
        )
        agent.add_argument(
            '--light_use_action',
            type=str,
            default='none',
            choices=['partner', 'self', 'all', 'none'],
        )
        agent.add_argument('--light_use_affordances', type='bool', default=False)
        agent.add_argument(
            '--light_use_current_self_output',
            type=str,
            default="none",
            choices=['none', 'all', 'all_filtered', 'all_filtered_remove'],
        )
        agent.add_argument(
            '--light_label_type',
            type=str,
            default='speech',
            choices=['speech', 'action', 'emote'],
            help='type of target in light dialogues',
        )
        agent.add_argument('--light_use_cands', type=int, default=20)
        agent.add_argument('--light_use_clip_cands', type=int, default=10000)
        agent.add_argument('--light_use_speech_prefix', type='bool', default=False)
        agent.add_argument('--light_use_score_cutoff', type=int, default=1)
        agent.add_argument('--light_use_max_score_cutoff', type=int, default=-1)
        agent.add_argument(
            '--light_use_hard_score_cutoff',
            type='bool',
            default=False,
            help='Specify True to **only** include examples with the specified '
            'score cutoff. E.g., if `--light-use-score-cutoff 3`, will only '
            'supply dialogues with scores of 3.',
        )
        agent.add_argument(
            '--light-model-name',
            type=str,
            default=None,
            help='if specified, the model from which chats should be saved',
        )
        agent.add_argument(
            '--light-use-continue-type',
            type=str,
            default='all',
            choices=['all', 'continue', 'exit'],
            help='only use dialogues that had a specific continue outcome, default use all',
        )
        agent.add_argument(
            '--light-use-date-cutoff',
            type=str,
            default=None,
            help=(
                'If specified, only include hobbot conversations that were collected '
                'before this date. Format YYYY-MM-DD.'
            ),
        )
        agent.add_argument(
            '--light-use-person-names-prefix',
            type='bool',
            default=False,
            help="If specified, prefix text with character name",
        )
        agent.add_argument('--light_percent_train_exs', type=float, default=1.0)

    def __init__(self, opt, shared=None):
        super().__init__(opt, shared)
        self.id += '_' + self.opt['light_label_type']


class SimpleMultiTeacher(DefaultTeacher):
    @staticmethod
    def add_cmdline_args(argparser):
        agent = argparser.add_argument_group('LIGHT Dialogue options')
        agent.add_argument(
            '--light_use_repeat',
            type=str,
            default='none',
            choices=['self_last', 'partner_last', 'none', 'both_last'],
        )
        agent.add_argument('--light_use_setting', type='bool', default=True)
        agent.add_argument('--light_use_unseen_test', type='bool', default=False)
        agent.add_argument('--light_use_person_names', type='bool', default=True)
        agent.add_argument(
            '--light_use_persona',
            type=str,
            default='self',
            choices=['partner', 'self', 'all', 'none'],
        )
        agent.add_argument('--light_use_taskname', type='bool', default=False)
        agent.add_argument(
            '--light_use_emote',
            type=str,
            default='none',
            choices=['partner', 'self', 'all', 'none'],
        )
        agent.add_argument(
            '--light_use_speech',
            type=str,
            default='partner',
            choices=['partner', 'self', 'all', 'none'],
        )
        agent.add_argument(
            '--light_use_action',
            type=str,
            default='none',
            choices=['partner', 'self', 'all', 'none'],
        )
        agent.add_argument('--light_use_affordances', type='bool', default=False)
        agent.add_argument(
            '--light_use_current_self_output',
            type=str,
            default="all",
            choices=['none', 'all', 'all_filtered', 'all_filtered_remove'],
        )
        agent.add_argument(
            '--light_label_type',
            type=str,
            default='speech',
            choices=['speech', 'action', 'emote', 'which'],
            help='type of target in light dialogues',
        )
        agent.add_argument('--light_use_cands', type=int, default=20)
        agent.add_argument('--light_use_clip_cands', type=int, default=10000)
        agent.add_argument('--light_use_speech_prefix', type='bool', default=False)
        agent.add_argument('--light_use_score_cutoff', type=int, default=1)
        agent.add_argument('--light_use_max_score_cutoff', type=int, default=-1)
        agent.add_argument(
            '--light_use_hard_score_cutoff',
            type='bool',
            default=False,
            help='Specify True to **only** include examples with the specified '
            'score cutoff. E.g., if `--light-use-score-cutoff 3`, will only '
            'supply dialogues with scores of 3.',
        )
        agent.add_argument(
            '--light-model-name',
            type=str,
            default=None,
            help='if specified, the model from which chats should be saved',
        )
        agent.add_argument(
            '--light-use-continue-type',
            type=str,
            default='all',
            choices=['all', 'continue', 'exit'],
            help='only use dialogues that had a specific continue outcome, default use all',
        )
        agent.add_argument(
            '--light-use-date-cutoff',
            type=str,
            default=None,
            help=(
                'If specified, only include hobbot conversations that were collected '
                'before this date. Format YYYY-MM-DD.'
            ),
        )
        agent.add_argument(
            '--light-use-person-names-prefix',
            type='bool',
            default=False,
            help="If specified, prefix text with character name",
        )
        agent.add_argument('--light_percent_train_exs', type=float, default=1.0)

    def __init__(self, opt, shared=None):
        super().__init__(opt, shared)
        self.id += '_' + self.opt['light_label_type']


class ReversedTeacher(DefaultTeacher):
    """
    Reversed teacher where labels are what bot said, text is what human said.

    Only support swapping text & labels, no other guarantees.
    """

    @staticmethod
    def add_cmdline_args(argparser):
        DefaultTeacher.add_cmdline_args(argparser)
        agent = argparser.add_argument_group('LIGHT Dialogue options')
        agent.add_argument(
            '--light-min-conv-length',
            type=int,
            default=None,
            help="if specified, filters out conversations that do not reach this length. "
            "`--light-min-conv-length 3` will filter out convos with less than 3 turns",
        )
        agent.add_argument(
            '--light-max-conv-length',
            type=int,
            default=None,
            help="if specified, caps conversations at this length. "
            "`--light-max-conv-length 6` will cap all conversations at 6 turns",
        )
        agent.add_argument(
            '--light-hard-max-length-cap',
            type='bool',
            default=False,
            help="if specificed, conversation max length is a hard cap - that is, "
            "any conversations greater than --light-max-conv-length will be "
            "removed completely",
        )
        agent.add_argument(
            '--light-prettify-setting',
            type='bool',
            default=False,
            help='if specified, setting/characters prettified.',
        )

    def _setup_data(self, path):
        super()._setup_data(path)
        new_eps = []
        min_len = self.opt.get('light_min_conv_length')
        if min_len:
            self.episodes = [e for e in self.episodes if len(e) >= min_len + 1]
        for ep in self.episodes:
            texts = [ex['text'] for ex in ep]
            labels = [ex['labels'] for ex in ep]
            new_ep = []
            context = texts[0]
            if self.opt.get('light_use_setting'):
                if self.opt.get('light_prettify_setting') and self.opt.get(
                    'light_use_person_names_prefix'
                ):
                    char1, char2 = [texts[1].split(':')[0], labels[0][0].split(':')[0]]
                    name, desc = context.split('_setting_desc')
                    name = name.replace('_setting_name', '').replace(', Somewhere', '')
                    context = (
                        f"A {char1} and a {char2} are in {name}. {desc.split('.')[0]}"
                    )
                new_ep = [
                    {'text': '', 'labels': [f"*{context}*"], 'episode_done': False}
                ]
            new_ep += [
                {'text': labels[i][0], 'labels': [texts[i + 1]], 'episode_done': False}
                for i in range(0, len(texts) - 1)
            ]
            if new_ep:
                new_ep[-1]['episode_done'] = True
                new_eps.append(new_ep)

        max_len = self.opt.get('light_max_conv_length')
        hard_max_cap = self.opt.get('light_hard_max_length_cap')
        if max_len:
            if hard_max_cap:
                new_eps = [e for e in new_eps if len(e) <= max_len]
            else:
                new_eps = [e[:max_len] for e in new_eps]
                for e in new_eps:
                    e[-1]['episode_done'] = True

        self.episodes = new_eps
        self.num_exs = sum([len(ep) for ep in self.episodes])


class SelfchatTeacher(SimpleTeacher):
    """
    Teacher used to create candidates for selfchats, if needed.
    """

    pass



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import copy
import os

from parlai.core.teachers import DialogTeacher
from parlai.tasks.multinli.agents import setup_data, BICLASS_LABELS, MULTINLI_LABELS

from .build import build

SNLI = 'SNLI'
SNLI_VERSION = '1.0'
SNLI_PREFIX = 'snli_'


def _path(opt):
    build(opt)

    dt = opt['datatype'].split(':')[0]

    if dt == 'train':
        suffix = 'train'
    elif dt == 'valid':
        suffix = 'dev'
    elif dt == 'test':
        suffix = 'test'
    else:
        raise RuntimeError('Not valid datatype.')

    data_path = os.path.join(
        opt['datapath'],
        SNLI,
        SNLI_PREFIX + SNLI_VERSION,
        SNLI_PREFIX + SNLI_VERSION + '_' + suffix + '.jsonl',
    )

    return data_path


class DefaultTeacher(DialogTeacher):
    @staticmethod
    def add_cmdline_args(parser):
        parser = parser.add_argument_group('SNLI Teacher Args')
        parser.add_argument(
            '-dfm',
            '--dialog-format',
            type='bool',
            default=False,
            help="True if one would like to convert to a dialogue format without special tokens such as 'Premise'"
            " and 'Hypothesis' (default: False).",
        )
        parser.add_argument(
            '-bcl',
            '--binary-classes',
            type='bool',
            default=False,
            help="True if label candidates are (contradiction, not_contradiction), and (entailment, contradiction, "
            "neutral) otherwise (default: False).",
        )

    def __init__(self, opt, shared=None):
        opt = copy.deepcopy(opt)
        data_path = _path(opt)
        opt['datafile'] = data_path
        self.id = 'SNLI'
        self.dialog_format = opt.get('dialog_format', False)
        self.binary_classes = opt.get('binary_classes', False)
        super().__init__(opt, shared)

    def setup_data(self, path):
        return setup_data(path, self.dialog_format, self.binary_classes)

    def label_candidates(self):
        if self.binary_classes:
            return BICLASS_LABELS
        return MULTINLI_LABELS



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import json
import logging
import os
import pickle
import sqlite3
import time
import threading

import parlai.mturk.core.shared_utils as shared_utils
from parlai.mturk.core.shared_utils import AssignState


def force_dir(path):
    """
    Make sure the parent dir exists for path so we can write a file.
    """
    os.makedirs(os.path.dirname(path), exist_ok=True)


data_dir = os.path.join(shared_utils.get_mturk_dir(), 'run_data')
os.makedirs(data_dir, exist_ok=True)

# Run data table:
CREATE_RUN_DATA_SQL_TABLE = """CREATE TABLE IF NOT EXISTS runs (
        run_id string PRIMARY KEY,
        created integer NOT NULL,
        maximum integer NOT NULL,
        completed integer NOT NULL,
        failed integer NOT NULL,
        taskname string,
        launch_time int
    );
    """

# Worker data table:
# TODO add block status
CREATE_WORKER_DATA_SQL_TABLE = """CREATE TABLE IF NOT EXISTS workers (
        worker_id string PRIMARY KEY,
        accepted integer NOT NULL,
        disconnected integer NOT NULL,
        expired integer NOT NULL,
        completed integer NOT NULL,
        approved integer NOT NULL,
        rejected integer NOT NULL
    );
    """

# HIT data table:
CREATE_HIT_DATA_SQL_TABLE = """CREATE TABLE IF NOT EXISTS hits (
        hit_id string PRIMARY KEY,
        expiration integer NOT NULL,
        hit_status string,
        assignments_pending int,
        assignments_available int,
        assignments_complete int,
        run_id string,
        FOREIGN KEY (run_id) REFERENCES runs (run_id)
    );
    """

# Assignment data table: (as one HIT can technically have multiple assignments)
CREATE_ASSIGN_DATA_SQL_TABLE = """CREATE TABLE IF NOT EXISTS assignments (
        assignment_id string PRIMARY KEY,
        status string,
        approve_time int,
        worker_id string,
        hit_id string,
        FOREIGN KEY (worker_id) REFERENCES workers (worker_id),
        FOREIGN KEY (hit_id) REFERENCES hits (hit_id)
    );
    """

# pairing data table: (reflects one worker<->assignment pairing)
CREATE_PAIRING_DATA_SQL_TABLE = """CREATE TABLE IF NOT EXISTS pairings (
        status string,
        onboarding_start int,
        onboarding_end int,
        task_start int,
        task_end int,
        conversation_id string,
        bonus_amount int,
        bonus_text string,
        bonus_paid boolean,
        notes string,
        worker_id string,
        assignment_id string,
        run_id string,
        onboarding_id string,
        extra_bonus_amount int,
        extra_bonus_text string,
        FOREIGN KEY (worker_id) REFERENCES workers (worker_id),
        FOREIGN KEY (assignment_id) REFERENCES assignments (assignment_id),
        FOREIGN KEY (run_id) REFERENCES runs (run_id)
    );
    """


class MTurkDataHandler:
    """
    Handles logging data to and reading data from a SQLite3 table for observation across
    processes and for controlled restarts.
    """

    def __init__(self, task_group_id=None, file_name='pmt_data.db'):
        self.db_path = os.path.join(data_dir, file_name)
        self.conn = {}
        self.task_group_id = task_group_id
        self.table_access_condition = threading.Condition()
        self.create_default_tables()

    @staticmethod
    def save_world_data(
        prepped_save_data, task_group_id, conversation_id, sandbox=False
    ):
        target = 'sandbox' if sandbox else 'live'
        if task_group_id is None:
            return
        target_dir = os.path.join(data_dir, target, task_group_id, conversation_id)
        custom_data = prepped_save_data['custom_data']
        if custom_data is not None:
            target_dir_custom = os.path.join(target_dir, 'custom')
            if custom_data.get('needs-pickle') is not None:
                pickle_file = os.path.join(target_dir_custom, 'data.pickle')
                force_dir(pickle_file)
                with open(pickle_file, 'wb') as outfile:
                    pickle.dump(custom_data['needs-pickle'], outfile)
                del custom_data['needs-pickle']
            custom_file = os.path.join(target_dir_custom, 'data.json')
            force_dir(custom_file)
            print('Saving data to {}.'.format(custom_file))
            with open(custom_file, 'w') as outfile:
                json.dump(custom_data, outfile)
        worker_data = prepped_save_data['worker_data']
        target_dir_workers = os.path.join(target_dir, 'workers')
        for worker_id, w_data in worker_data.items():
            worker_file = os.path.join(target_dir_workers, '{}.json'.format(worker_id))
            force_dir(worker_file)
            with open(worker_file, 'w') as outfile:
                json.dump(w_data, outfile)

    def _get_connection(self):
        """
        Returns a singular database connection to be shared amongst all calls.
        """
        curr_thread = threading.get_ident()
        if curr_thread not in self.conn or self.conn[curr_thread] is None:
            try:
                conn = sqlite3.connect(self.db_path)
                conn.row_factory = sqlite3.Row
                self.conn[curr_thread] = conn
            except sqlite3.Error as e:
                shared_utils.print_and_log(
                    logging.ERROR,
                    "Could not get db connection, failing: {}".format(repr(e)),
                    should_print=True,
                )
                raise e
        return self.conn[curr_thread]

    def _force_task_group_id(self, task_group_id):
        """
        Throw an error if a task group id is neither provided nor stored.
        """
        if task_group_id is None:
            task_group_id = self.task_group_id
        assert task_group_id is not None, 'Default task_group_id not set'
        return task_group_id

    def create_default_tables(self):
        """
        Prepares the default tables in the database if they don't exist.
        """
        with self.table_access_condition:
            conn = self._get_connection()
            c = conn.cursor()
            c.execute(CREATE_RUN_DATA_SQL_TABLE)
            c.execute(CREATE_WORKER_DATA_SQL_TABLE)
            c.execute(CREATE_HIT_DATA_SQL_TABLE)
            c.execute(CREATE_ASSIGN_DATA_SQL_TABLE)
            c.execute(CREATE_PAIRING_DATA_SQL_TABLE)

            # TODO move to versioning and migrations
            # attempt to add new fields one by one to update old versions
            updates = [
                '''ALTER TABLE pairings
                   ADD COLUMN onboarding_id TEXT default null''',
                '''ALTER TABLE runs
                   ADD COLUMN taskname TEXT default null''',
                '''ALTER TABLE runs
                   ADD COLUMN launch_time int default 0''',
                '''ALTER TABLE pairings
                   ADD COLUMN extra_bonus_amount INT default 0''',
                '''ALTER TABLE pairings
                   ADD COLUMN extra_bonus_text TEXT default '';''',
            ]
            for update in updates:
                try:
                    c.execute(update)
                except sqlite3.Error:
                    pass
            conn.commit()

    def log_new_run(self, target_hits, taskname, task_group_id=None):
        """
        Add a new run to the runs table.
        """
        with self.table_access_condition:
            task_group_id = self._force_task_group_id(task_group_id)
            conn = self._get_connection()
            c = conn.cursor()
            c.execute(
                'INSERT INTO runs VALUES (?,?,?,?,?,?,?);',
                (task_group_id, 0, target_hits, 0, 0, taskname, time.time()),
            )
            conn.commit()

    def log_hit_status(self, mturk_hit_creation_response, task_group_id=None):
        """
        Create or update an entry in the hit status table.
        """
        task_group_id = self._force_task_group_id(task_group_id)

        hit_details = mturk_hit_creation_response['HIT']
        id = hit_details['HITId']
        expiration = time.mktime(hit_details['Expiration'].timetuple())
        status = hit_details['HITStatus']
        assignments_pending = hit_details['NumberOfAssignmentsPending']
        assignments_available = hit_details['NumberOfAssignmentsAvailable']
        assignments_complete = hit_details['NumberOfAssignmentsCompleted']
        with self.table_access_condition:
            conn = self._get_connection()
            c = conn.cursor()
            c.execute('SELECT COUNT(*) FROM hits WHERE hit_id = ?;', (id,))
            is_new_hit = c.fetchone()[0] == 0
            if is_new_hit:
                c.execute(
                    """UPDATE runs SET created = created + 1
                             WHERE run_id = ?;""",
                    (task_group_id,),
                )

            c.execute(
                'REPLACE INTO hits VALUES (?,?,?,?,?,?,?);',
                (
                    id,
                    expiration,
                    status,
                    assignments_pending,
                    assignments_available,
                    assignments_complete,
                    task_group_id,
                ),
            )
            conn.commit()

    def log_worker_accept_assignment(
        self, worker_id, assignment_id, hit_id, task_group_id=None
    ):
        """
        Log a worker accept, update assignment state and pairings to match the
        acceptance.
        """
        task_group_id = self._force_task_group_id(task_group_id)
        with self.table_access_condition:
            conn = self._get_connection()
            c = conn.cursor()

            # Ensure worker exists, mark the accepted assignment
            c.execute('SELECT COUNT(*) FROM workers WHERE worker_id = ?;', (worker_id,))
            has_worker = c.fetchone()[0] > 0
            if not has_worker:
                # Must instert a new worker into the database
                c.execute(
                    'INSERT INTO workers VALUES (?,?,?,?,?,?,?);',
                    (worker_id, 1, 0, 0, 0, 0, 0),
                )
            else:
                # Increment number of assignments the worker has accepted
                c.execute(
                    """UPDATE workers SET accepted = accepted + 1
                             WHERE worker_id = ?;""",
                    (worker_id,),
                )

            # Ensure the assignment exists, mark the current worker
            c.execute(
                'REPLACE INTO assignments VALUES (?,?,?,?,?)',
                (assignment_id, 'Accepted', None, worker_id, hit_id),
            )

            # Create tracking for this specific pairing, as the assignment
            # may be reassigned
            c.execute(
                """INSERT INTO pairings
                         VALUES (?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?)""",
                (
                    AssignState.STATUS_NONE,
                    None,
                    None,
                    None,
                    None,
                    None,
                    0,
                    '',
                    False,
                    '',
                    worker_id,
                    assignment_id,
                    task_group_id,
                    None,
                    0,
                    '',
                ),
            )
            conn.commit()

    def log_complete_assignment(
        self, worker_id, assignment_id, approve_time, complete_type, task_group_id=None
    ):
        """
        Note that an assignment was completed.
        """
        task_group_id = self._force_task_group_id(task_group_id)
        with self.table_access_condition:
            conn = self._get_connection()
            c = conn.cursor()
            # Update assign data to completed
            c.execute(
                """UPDATE assignments SET status = ?, approve_time = ?
                         WHERE assignment_id = ?;""",
                ('Completed', approve_time, assignment_id),
            )

            # Increment worker completed
            c.execute(
                """UPDATE workers SET completed = completed + 1
                         WHERE worker_id = ?;""",
                (worker_id,),
            )

            # update the payment data status
            c.execute(
                """UPDATE pairings SET status = ?, task_end = ?
                         WHERE worker_id = ? AND assignment_id = ?;""",
                (complete_type, time.time(), worker_id, assignment_id),
            )

            # Update run data to have another completed
            c.execute(
                """UPDATE runs SET completed = completed + 1
                         WHERE run_id = ?;""",
                (task_group_id,),
            )
            conn.commit()

    def log_disconnect_assignment(
        self,
        worker_id,
        assignment_id,
        approve_time,
        disconnect_type,
        task_group_id=None,
    ):
        """
        Note that an assignment was disconnected from.
        """
        task_group_id = self._force_task_group_id(task_group_id)
        with self.table_access_condition:
            conn = self._get_connection()
            c = conn.cursor()

            # Update assign data to completed for this task (we can't track)
            c.execute(
                """UPDATE assignments SET status = ?, approve_time = ?
                         WHERE assignment_id = ?;""",
                ('Disconnected', approve_time, assignment_id),
            )

            # Increment worker disconnected
            c.execute(
                """UPDATE workers SET disconnected = disconnected + 1
                         WHERE worker_id = ?;""",
                (worker_id,),
            )

            # update the pairing status
            c.execute(
                """UPDATE pairings SET status = ?, task_end = ?
                         WHERE worker_id = ? AND assignment_id = ?;""",
                (disconnect_type, time.time(), worker_id, assignment_id),
            )

            # Update run data to have another completed
            c.execute(
                'UPDATE runs SET failed = failed + 1 WHERE run_id = ?;',
                (task_group_id,),
            )
            conn.commit()

    def log_expire_assignment(self, worker_id, assignment_id, task_group_id=None):
        """
        Note that an assignment was expired by us.
        """
        task_group_id = self._force_task_group_id(task_group_id)
        with self.table_access_condition:
            conn = self._get_connection()
            c = conn.cursor()

            # Update assign data to expired
            c.execute(
                """UPDATE assignments SET status = ?
                         WHERE assignment_id = ?;""",
                ('Expired', assignment_id),
            )

            # Increment worker completed
            c.execute(
                """UPDATE workers SET expired = expired + 1
                         WHERE worker_id = ?;""",
                (worker_id,),
            )

            # update the pairing status
            c.execute(
                """UPDATE pairings SET status = ?, task_end = ?
                         WHERE worker_id = ? AND assignment_id = ?;""",
                (AssignState.STATUS_EXPIRED, time.time(), worker_id, assignment_id),
            )

            # Update run data to have another completed
            c.execute(
                'UPDATE runs SET failed = failed + 1 WHERE run_id = ?;',
                (task_group_id,),
            )
            conn.commit()

    def log_submit_assignment(self, worker_id, assignment_id):
        """
        To be called whenever a worker hits the "submit hit" button.
        """
        with self.table_access_condition:
            conn = self._get_connection()
            c = conn.cursor()
            # update the assignment status to reviewable
            c.execute(
                """UPDATE assignments SET status = ?
                         WHERE assignment_id = ?;""",
                ('Reviewable', assignment_id),
            )
            conn.commit()

    def log_abandon_assignment(self, worker_id, assignment_id):
        """
        To be called whenever a worker returns a hit.
        """
        with self.table_access_condition:
            conn = self._get_connection()
            c = conn.cursor()
            # update the assignment status to reviewable
            c.execute(
                """UPDATE assignments SET status = ?
                         WHERE assignment_id = ?;""",
                ('Abandoned', assignment_id),
            )
            # Increment worker completed
            c.execute(
                """UPDATE workers SET disconnected = disconnected + 1
                         WHERE worker_id = ?;""",
                (worker_id,),
            )
            conn.commit()

    def log_start_onboard(self, worker_id, assignment_id, conversation_id):
        """
        Update a pairing state to reflect onboarding status.
        """
        with self.table_access_condition:
            conn = self._get_connection()
            c = conn.cursor()
            c.execute(
                '''UPDATE pairings SET status = ?, onboarding_start = ?,
                         onboarding_id = ?
                         WHERE worker_id = ? AND assignment_id = ?;''',
                (
                    AssignState.STATUS_ONBOARDING,
                    time.time(),
                    conversation_id,
                    worker_id,
                    assignment_id,
                ),
            )
            conn.commit()

    def log_finish_onboard(self, worker_id, assignment_id):
        """
        Update a pairing state to reflect waiting status.
        """
        with self.table_access_condition:
            conn = self._get_connection()
            c = conn.cursor()
            c.execute(
                """UPDATE pairings SET status = ?, onboarding_end = ?
                         WHERE worker_id = ? AND assignment_id = ?;""",
                (AssignState.STATUS_WAITING, time.time(), worker_id, assignment_id),
            )
            conn.commit()

    def log_start_task(self, worker_id, assignment_id, conversation_id):
        """
        Update a pairing state to reflect in_task status.
        """
        with self.table_access_condition:
            conn = self._get_connection()
            c = conn.cursor()
            c.execute(
                """UPDATE pairings SET status = ?, task_start = ?,
                         conversation_id = ? WHERE worker_id = ?
                         AND assignment_id = ?;""",
                (
                    AssignState.STATUS_IN_TASK,
                    time.time(),
                    conversation_id,
                    worker_id,
                    assignment_id,
                ),
            )
            conn.commit()

    def log_award_amount(self, worker_id, assignment_id, amount, reason):
        """
        Update a pairing state to add a task bonus to be paid, appends reason.

        To be used for automatic evaluation bonuses
        """
        with self.table_access_condition:
            conn = self._get_connection()
            c = conn.cursor()
            reason = "${} for {}\n".format(amount, reason)
            # Bonus amount is stored in a cents int in the SQL table
            cent_amount = int(amount * 100)
            c.execute(
                """UPDATE pairings SET bonus_amount = bonus_amount + ?,
                        bonus_text = bonus_text || ?
                         WHERE worker_id = ? AND assignment_id = ?;""",
                (cent_amount, reason, worker_id, assignment_id),
            )
            conn.commit()

    def log_pay_extra_bonus(self, worker_id, assignment_id, amount, reason):
        """
        Update a pairing state to add a bonus to be paid, appends reason.

        To be used for extra bonuses awarded at the discretion of the requester
        """
        with self.table_access_condition:
            conn = self._get_connection()
            c = conn.cursor()
            reason = "${} for {}\n".format(amount, reason)
            # Bonus amount is stored in a cents int in the SQL table
            cent_amount = int(amount * 100)
            c.execute(
                """UPDATE pairings
                        SET extra_bonus_amount = extra_bonus_amount + ?,
                        extra_bonus_text = extra_bonus_text || ?
                         WHERE worker_id = ? AND assignment_id = ?;""",
                (cent_amount, reason, worker_id, assignment_id),
            )
            conn.commit()

    def log_bonus_paid(self, worker_id, assignment_id):
        """
        Update to show that the intended bonus amount awarded for work in the task has
        been paid.
        """
        with self.table_access_condition:
            conn = self._get_connection()
            c = conn.cursor()
            c.execute(
                """UPDATE pairings SET bonus_paid = ?
                         WHERE worker_id = ? AND assignment_id = ?;""",
                (True, worker_id, assignment_id),
            )
            conn.commit()

    def log_approve_assignment(self, assignment_id):
        """
        Update assignment state to reflect approval, update worker state to increment
        number of accepted assignments.
        """
        with self.table_access_condition:
            conn = self._get_connection()
            c = conn.cursor()

            c.execute(
                'SELECT * FROM assignments WHERE assignment_id = ?;', (assignment_id,)
            )
            assignment = c.fetchone()
            if assignment is None:
                return
            status = assignment['status']
            worker_id = assignment['worker_id']
            if status == 'Approved':
                return  # assign already approved

            # handle the approving part
            c.execute(
                """UPDATE assignments SET status = ?
                         WHERE assignment_id = ?;""",
                ('Approved', assignment_id),
            )

            # This was a rejection override
            if status == 'Rejected':
                c.execute(
                    """UPDATE workers SET approved = approved + 1,
                             rejected = rejected - 1
                             WHERE worker_id = ?;""",
                    (worker_id,),
                )
            else:
                c.execute(
                    """UPDATE workers SET approved = approved + 1
                             WHERE worker_id = ?;""",
                    (worker_id,),
                )

            conn.commit()

    def log_reject_assignment(self, assignment_id):
        """
        Update assignment state to reflect rejection, update worker state to increment
        number of rejected assignments.
        """
        with self.table_access_condition:
            conn = self._get_connection()
            c = conn.cursor()
            c.execute(
                """UPDATE assignments SET status = ?
                         WHERE assignment_id = ? AND status != ?;""",
                ('Rejected', assignment_id, 'Rejected'),
            )
            # If this reject actually happened, update worker's rejections
            if c.rowcount > 0:
                c.execute(
                    'SELECT * FROM assignments WHERE assignment_id = ?;',
                    (assignment_id,),
                )
                assignment = c.fetchone()
                if assignment is None:
                    return
                worker_id = assignment['worker_id']
                c.execute(
                    """UPDATE workers SET rejected = rejected + 1
                             WHERE worker_id = ?;""",
                    (worker_id,),
                )
            conn.commit()

    def log_worker_note(self, worker_id, assignment_id, note):
        """
        Append a note to the worker notes for a particular worker-assignment pairing.

        Adds newline to the note.
        """
        note += '\n'
        with self.table_access_condition:
            try:
                conn = self._get_connection()
                c = conn.cursor()
                c.execute(
                    """UPDATE pairings SET notes = notes || ?
                             WHERE worker_id = ? AND assignment_id = ?;""",
                    (note, worker_id, assignment_id),
                )
                conn.commit()
            except Exception as e:
                print(repr(e))

    def get_all_worker_data(self, start=0, count=100):
        """
        get all the worker data for all worker_ids.
        """
        with self.table_access_condition:
            conn = self._get_connection()
            c = conn.cursor()
            c.execute('SELECT * FROM workers LIMIT ?,?;', (start, start + count))
            results = c.fetchall()
            return results

    def get_worker_data(self, worker_id):
        """
        get all worker data for a particular worker_id.
        """
        with self.table_access_condition:
            conn = self._get_connection()
            c = conn.cursor()
            c.execute('SELECT * FROM workers WHERE worker_id = ?;', (worker_id,))
            results = c.fetchone()
            return results

    def get_assignments_for_run(self, task_group_id):
        """
        get all assignments for a particular run by task_group_id.
        """
        with self.table_access_condition:
            conn = self._get_connection()
            c = conn.cursor()
            c.execute(
                """SELECT assignments.* FROM assignments
                         WHERE assignments.hit_id IN (
                           SELECT hits.hit_id FROM hits
                           WHERE hits.run_id = ?
                         );""",
                (task_group_id,),
            )
            results = c.fetchall()
            return results

    def get_assignment_data(self, assignment_id):
        """
        get assignment data for a particular assignment by assignment_id.
        """
        with self.table_access_condition:
            conn = self._get_connection()
            c = conn.cursor()
            c.execute(
                'SELECT * FROM assignments WHERE assignment_id = ?;', (assignment_id,)
            )
            results = c.fetchone()
            return results

    def get_worker_assignment_pairing(self, worker_id, assignment_id):
        """
        get a pairing data structure between a worker and an assignment.
        """
        with self.table_access_condition:
            conn = self._get_connection()
            c = conn.cursor()
            c.execute(
                """SELECT * FROM pairings WHERE worker_id = ?
                         AND assignment_id = ?;""",
                (worker_id, assignment_id),
            )
            results = c.fetchone()
            return results

    def get_all_run_data(self, start=0, count=1000):
        """
        get all the run data for all task_group_ids.
        """
        with self.table_access_condition:
            conn = self._get_connection()
            c = conn.cursor()
            c.execute('SELECT * FROM runs LIMIT ?,?;', (start, start + count))
            results = c.fetchall()
            return results

    def get_run_data(self, task_group_id):
        """
        get the run data for the given task_group_id, return None if not found.
        """
        with self.table_access_condition:
            conn = self._get_connection()
            c = conn.cursor()
            c.execute('SELECT * FROM runs WHERE run_id = ?;', (task_group_id,))
            results = c.fetchone()
            return results

    def get_hits_for_run(self, run_id):
        """
        Get the full list of HITs for the given run_id.
        """
        with self.table_access_condition:
            conn = self._get_connection()
            c = conn.cursor()
            c.execute("SELECT * FROM hits WHERE run_id = ?;", (run_id,))
            results = c.fetchall()
            return results

    def get_hit_data(self, hit_id):
        """
        get the hit data for the given hit_id, return None if not.
        """
        with self.table_access_condition:
            conn = self._get_connection()
            c = conn.cursor()
            c.execute("SELECT * FROM hits WHERE hit_id = ?;", (hit_id,))
            results = c.fetchone()
            return results

    def get_pairings_for_assignment(self, assignment_id):
        """
        get all pairings attached to a particular assignment_id.
        """
        with self.table_access_condition:
            conn = self._get_connection()
            c = conn.cursor()
            c.execute(
                "SELECT * FROM pairings WHERE assignment_id = ?;", (assignment_id,)
            )
            results = c.fetchall()
            return results

    def get_pairings_for_run(self, task_group_id):
        """
        get all pairings from a particular run by task_group_id.
        """
        with self.table_access_condition:
            conn = self._get_connection()
            c = conn.cursor()
            c.execute(
                """SELECT * FROM pairings
                         WHERE run_id = ?;""",
                (task_group_id,),
            )
            results = c.fetchall()
            return results

    def get_pairings_for_conversation(self, conversation_id, task_group_id=None):
        """
        get all pairings for a singular conversation in a run by conversation_id and
        task_group_id.
        """
        task_group_id = self._force_task_group_id(task_group_id)
        with self.table_access_condition:
            conn = self._get_connection()
            c = conn.cursor()
            c.execute(
                """SELECT * FROM pairings WHERE conversation_id = ?
                         AND run_id = ?;""",
                (conversation_id, task_group_id),
            )
            results = c.fetchall()
            return results

    def get_all_assignments_for_worker(self, worker_id):
        """
        get all assignments associated with a particular worker_id.
        """
        with self.table_access_condition:
            conn = self._get_connection()
            c = conn.cursor()
            c.execute("SELECT * FROM assignments WHERE worker_id = ?;", (worker_id,))
            results = c.fetchall()
            return results

    def get_all_pairings_for_worker(self, worker_id):
        """
        get all pairings associated with a particular worker_id.
        """
        with self.table_access_condition:
            conn = self._get_connection()
            c = conn.cursor()
            c.execute('SELECT * FROM pairings WHERE worker_id = ?;', (worker_id,))
            results = c.fetchall()
            return results

    def get_all_task_assignments_for_worker(self, worker_id, task_group_id=None):
        """
        get all assignments for a particular worker within a particular run by worker_id
        and task_group_id.
        """
        task_group_id = self._force_task_group_id(task_group_id)
        with self.table_access_condition:
            conn = self._get_connection()
            c = conn.cursor()
            c.execute(
                """SELECT assignments.assignment_id, assignments.status,
                         assignments.approve_time, assignments.worker_id,
                         assignments.hit_id
                         FROM assignments
                         INNER JOIN hits on assignments.hit_id = hits.hit_id
                         WHERE assignments.worker_id = ? AND hits.run_id = ?;
                         """,
                (worker_id, task_group_id),
            )
            results = c.fetchall()
            return results

    def get_all_task_pairings_for_worker(self, worker_id, task_group_id=None):
        """
        get all pairings for a particular worker within a particular run by worker_id
        and task_group_id.
        """
        task_group_id = self._force_task_group_id(task_group_id)
        with self.table_access_condition:
            conn = self._get_connection()
            c = conn.cursor()
            c.execute(
                """SELECT * FROM pairings WHERE worker_id = ?
                         AND run_id = ?;""",
                (worker_id, task_group_id),
            )
            results = c.fetchall()
            return results

    @staticmethod
    def get_conversation_data(task_group_id, conv_id, worker_id, is_sandbox):
        """
        A poorly named function that gets conversation data for a worker.
        """
        result = {
            'had_data_dir': False,
            'had_run_dir': False,
            'had_conversation_dir': False,
            'had_worker_dir': False,
            'had_worker_file': False,
            'data': None,
        }
        target = 'sandbox' if is_sandbox else 'live'
        search_dir = os.path.join(data_dir, target)
        if not os.path.exists(search_dir):
            return result
        result['had_data_dir'] = True
        search_dir = os.path.join(search_dir, task_group_id)
        if not os.path.exists(search_dir):
            return result
        result['had_run_dir'] = True
        search_dir = os.path.join(search_dir, conv_id)
        if not os.path.exists(search_dir):
            return result
        result['had_conversation_dir'] = True
        search_dir = os.path.join(search_dir, 'workers')
        if not os.path.exists(search_dir):
            return result
        result['had_worker_dir'] = True
        target_filename = os.path.join(search_dir, '{}.json'.format(worker_id))
        if not os.path.exists(target_filename):
            return result
        result['had_worker_file'] = True
        with open(target_filename, 'r') as target_file:
            result['data'] = json.load(target_file)

        return result

    @staticmethod
    def get_full_conversation_data(task_group_id, conv_id, is_sandbox):
        """
        Gets all conversation data saved for a world.
        """
        target = 'sandbox' if is_sandbox else 'live'
        return_data = {'custom_data': {}, 'worker_data': {}}

        target_dir = os.path.join(data_dir, target, task_group_id, conv_id)
        target_dir_custom = os.path.join(target_dir, 'custom')
        custom_file = os.path.join(target_dir_custom, 'data.json')
        if os.path.exists(custom_file):
            custom_data = {}
            with open(custom_file, 'r') as infile:
                custom_data = json.load(infile)
            pickle_file = os.path.join(target_dir_custom, 'data.pickle')
            if os.path.exists(pickle_file):
                with open(pickle_file, 'rb') as infile:
                    custom_data['needs-pickle'] = pickle.load(infile)
            return_data['custom_data'] = custom_data

        target_dir_workers = os.path.join(target_dir, 'workers')
        for w_file in os.listdir(target_dir_workers):
            w_id = w_file.split('.json')[0]
            worker_file = os.path.join(target_dir_workers, w_file)
            with open(worker_file, 'r') as infile:
                return_data['worker_data'][w_id] = json.load(infile)

        return return_data



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
from parlai.core.params import ParlaiParser
from parlai.core.agents import create_agent
from parlai.utils.misc import AttrDict
from parlai.mturk.core.mturk_manager import MTurkManager
import parlai.mturk.core.mturk_utils as mturk_utils

from worlds import WizardEval, TopicsGenerator, TopicChooseWorld
from task_config import task_config

from projects.wizard_of_wikipedia.knowledge_retriever.knowledge_retriever import (
    KnowledgeRetrieverAgent,
)

import gc
import datetime
import json
import os
import sys
from parlai.utils.logging import ParlaiLogger, INFO

MASTER_QUALIF = {
    'QualificationTypeId': '2F1QJWKUDD8XADTFD2Q0G6UTO95ALH',
    'Comparator': 'Exists',
    'RequiredToPreview': True,
}


def main():
    """
    This task consists of an MTurk agent evaluating a wizard model.

    They are assigned a topic and asked to chat.
    """
    start_time = datetime.datetime.today().strftime('%Y-%m-%d-%H-%M')
    argparser = ParlaiParser(False, add_model_args=True)
    argparser.add_parlai_data_path()
    argparser.add_mturk_args()
    argparser.add_argument(
        '-mt', '--max-turns', default=10, type=int, help='maximal number of chat turns'
    )
    argparser.add_argument(
        '--max-resp-time',
        default=240,
        type=int,
        help='time limit for entering a dialog message',
    )
    argparser.add_argument(
        '--generative-setup',
        default=False,
        help='mimic setup for the WoW generator task (use knowledge token)',
    )
    argparser.add_argument(
        '--max-choice-time',
        type=int,
        default=300,
        help='time limit for turker' 'choosing the topic',
    )
    argparser.add_argument(
        '--ag-shutdown-time',
        default=120,
        type=int,
        help='time limit for entering a dialog message',
    )
    argparser.add_argument(
        '-rt', '--range-turn', default='3,5', help='sample range of number of turns'
    )
    argparser.add_argument(
        '--human-eval',
        type='bool',
        default=False,
        help='human vs human eval, no models involved',
    )
    argparser.add_argument(
        '--auto-approve-delay',
        type=int,
        default=3600 * 24 * 1,
        help='how long to wait for auto approval',
    )
    argparser.add_argument(
        '--only-masters',
        type='bool',
        default=False,
        help='Set to true to use only master turks for ' 'this test eval',
    )
    argparser.add_argument(
        '--unique-workers',
        type='bool',
        default=False,
        help='Each worker must be unique',
    )
    argparser.add_argument(
        '--prepend-gold-knowledge',
        type='bool',
        default=False,
        help='Add the gold knowledge to the input text from the human for '
        'the model observation.',
    )
    argparser.add_argument(
        '--mturk-log',
        type=str,
        default='data/mturklogs/wizard_of_wikipedia/{}.log'.format(start_time),
    )

    def inject_override(opt, override_dict):
        opt['override'] = override_dict
        for k, v in override_dict.items():
            opt[k] = v

    def get_logger(opt):
        fmt = '%(asctime)s: [ %(message)s ]'
        logfile = None
        if 'mturk_log' in opt:
            logfile = opt['mturk_log']
            if not os.path.isdir(os.path.dirname(logfile)):
                os.makedirs(os.path.dirname(logfile))
        logger = ParlaiLogger(
            "mturk_woz",
            console_level=INFO,
            file_level=INFO,
            console_format=fmt,
            file_format=fmt,
            filename=logfile,
        )
        logger.info('COMMAND: %s' % ' '.join(sys.argv))
        logger.info('-' * 100)
        logger.info('CONFIG:\n%s' % json.dumps(opt, indent=4, sort_keys=True))

        return logger

    # MODEL CONFIG
    # NOTE: please edit this to test your own models
    config = {
        'model_file': 'models:wizard_of_wikipedia/end2end_generator/model',
        'generative_setup': True,
        'prepend_gold_knowledge': True,
        'model': 'projects:wizard_of_wikipedia:generator',
        'beam_size': 10,  # add inference arguments here
        'inference': 'beam',
        'beam_block_ngram': 3,
    }

    # add dialogue model args
    argparser.add_model_subargs(config['model'])
    # add knowledge retriever args
    argparser.add_model_subargs('projects:wizard_of_wikipedia:knowledge_retriever')
    start_opt = argparser.parse_args()

    inject_override(start_opt, config)

    if not start_opt.get('human_eval'):
        # make dialogue responder model
        bot = create_agent(start_opt)
        shared_bot_params = bot.share()
        # make knowledge retriever
        knowledge_opt = {
            'model': 'projects:wizard_of_wikipedia:knowledge_retriever',
            'add_token_knowledge': not start_opt['generative_setup'],
            'datapath': start_opt['datapath'],
            'interactive_mode': False,  # interactive mode automatically sets fixed cands
        }
        # add all existing opt to the knowledge opt, without overriding
        # the above arguments
        for k, v in start_opt.items():
            if k not in knowledge_opt and k not in config:
                knowledge_opt[k] = v

        knowledge_agent = KnowledgeRetrieverAgent(knowledge_opt)
        knowledge_agent_shared_params = knowledge_agent.share()

    else:
        shared_bot_params = None

    if not start_opt['human_eval']:
        get_logger(bot.opt)
    else:
        get_logger(start_opt)

    if start_opt['human_eval']:
        folder_name = 'human_eval-{}'.format(start_time)
    else:
        folder_name = '{}-{}'.format(start_opt['model'], start_time)

    start_opt['task'] = os.path.basename(os.path.dirname(os.path.abspath(__file__)))
    if 'data_path' not in start_opt:
        start_opt['data_path'] = os.path.join(
            os.getcwd(), 'data', 'wizard_eval', folder_name
        )
    start_opt.update(task_config)

    if not start_opt.get('human_eval'):
        mturk_agent_ids = ['PERSON_1']
    else:
        mturk_agent_ids = ['PERSON_1', 'PERSON_2']

    mturk_manager = MTurkManager(opt=start_opt, mturk_agent_ids=mturk_agent_ids)

    topics_generator = TopicsGenerator(start_opt)
    directory_path = os.path.dirname(os.path.abspath(__file__))
    mturk_manager.setup_server(task_directory_path=directory_path)
    worker_roles = {}
    connect_counter = AttrDict(value=0)

    try:
        mturk_manager.start_new_run()
        agent_qualifications = []
        if not start_opt['is_sandbox']:
            # assign qualifications
            if start_opt['only_masters']:
                agent_qualifications.append(MASTER_QUALIF)
            if start_opt['unique_workers']:
                qual_name = 'UniqueChatEval'
                qual_desc = (
                    'Qualification to ensure each worker completes a maximum '
                    'of one of these chat/eval HITs'
                )
                qualification_id = mturk_utils.find_or_create_qualification(
                    qual_name, qual_desc, False
                )
                print('Created qualification: ', qualification_id)
                UNIQUE_QUALIF = {
                    'QualificationTypeId': qualification_id,
                    'Comparator': 'DoesNotExist',
                    'RequiredToPreview': True,
                }
                start_opt['unique_qualif_id'] = qualification_id
                agent_qualifications.append(UNIQUE_QUALIF)
        mturk_manager.create_hits(qualifications=agent_qualifications)

        def run_onboard(worker):
            if start_opt['human_eval']:
                role = mturk_agent_ids[connect_counter.value % len(mturk_agent_ids)]
                connect_counter.value += 1
                worker_roles[worker.worker_id] = role
            else:
                role = 'PERSON_1'
            worker.topics_generator = topics_generator
            world = TopicChooseWorld(start_opt, worker, role=role)
            world.parley()
            world.shutdown()

        mturk_manager.set_onboard_function(onboard_function=run_onboard)
        mturk_manager.ready_to_accept_workers()

        def check_single_worker_eligibility(worker):
            return True

        def check_multiple_workers_eligibility(workers):
            valid_workers = {}
            for worker in workers:
                worker_id = worker.worker_id
                if worker_id not in worker_roles:
                    print('Something went wrong')
                    continue
                role = worker_roles[worker_id]
                if role not in valid_workers:
                    valid_workers[role] = worker
                if len(valid_workers) == 2:
                    break
            return valid_workers.values() if len(valid_workers) == 2 else []

        if not start_opt['human_eval']:
            eligibility_function = {
                'func': check_single_worker_eligibility,
                'multiple': False,
            }
        else:
            eligibility_function = {
                'func': check_multiple_workers_eligibility,
                'multiple': True,
            }

        def assign_worker_roles(workers):
            if start_opt['human_eval']:
                for worker in workers:
                    worker.id = worker_roles[worker.worker_id]
            else:
                for index, worker in enumerate(workers):
                    worker.id = mturk_agent_ids[index % len(mturk_agent_ids)]

        def run_conversation(mturk_manager, opt, workers):
            conv_idx = mturk_manager.conversation_index
            world = WizardEval(
                opt=start_opt,
                agents=workers,
                range_turn=[int(s) for s in start_opt['range_turn'].split(',')],
                max_turn=start_opt['max_turns'],
                max_resp_time=start_opt['max_resp_time'],
                model_agent_opt=shared_bot_params,
                world_tag='conversation t_{}'.format(conv_idx),
                agent_timeout_shutdown=opt['ag_shutdown_time'],
                knowledge_retriever_opt=knowledge_agent_shared_params,
            )
            while not world.episode_done():
                world.parley()
            world.save_data()

            world.shutdown()
            gc.collect()

        mturk_manager.start_task(
            eligibility_function=eligibility_function,
            assign_role_function=assign_worker_roles,
            task_function=run_conversation,
        )

    except BaseException:
        raise
    finally:
        mturk_manager.expire_all_unassigned_hits()
        mturk_manager.shutdown()


if __name__ == '__main__':
    main()



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import errno
import logging
import json
import threading
import time
from queue import PriorityQueue, Empty
import websocket

from parlai.mturk.core.dev.shared_utils import print_and_log
import parlai.mturk.core.dev.data_model as data_model
import parlai.mturk.core.dev.shared_utils as shared_utils


class Packet:
    """
    Class for holding information sent over a socket.
    """

    # Possible Packet Status
    STATUS_NONE = -1
    STATUS_INIT = 0
    STATUS_SENT = 1
    STATUS_FAIL = 2

    # TODO remove unused attributes
    def __init__(
        self,
        id,
        type,
        sender_id,
        receiver_id,
        assignment_id,
        data,
        conversation_id=None,
        requires_ack=None,
        blocking=None,
        ack_func=None,
    ):
        """
        Create a packet to be used for holding information before it is
        sent through the socket
        id:               Unique ID to distinguish this packet from others
        type:             TYPE of packet (ACK, ALIVE, MESSAGE)
        sender_id:        Sender ID for this packet
        receiver_id:      Recipient ID for this packet
        assignment_id:    Assignment ID for this packet
        data:             Contents of the packet
        conversation_id:  Packet metadata - what conversation this belongs to
        requires_ack:     No longer used.
        blocking:         No longer used.
        ack_func:         Function to call upon successful ack of a packet
                           Default calls no function on ack
        """
        self.id = id
        # Possible Packet Types are set by data_model
        self.type = type
        self.sender_id = sender_id
        self.receiver_id = receiver_id
        self.assignment_id = assignment_id
        self.data = data
        self.conversation_id = conversation_id
        self.ack_func = ack_func
        self.status = self.STATUS_INIT

    @staticmethod
    def from_dict(packet):
        """
        Create a packet from the dictionary that would be recieved over a socket.
        """
        try:
            packet_id = packet['id']
            packet_type = packet['type']
            sender_id = packet['sender_id']
            receiver_id = packet.get('receiver_id', None)
            assignment_id = packet.get('assignment_id', None)
            data = packet.get('data', '')
            conversation_id = packet.get('conversation_id', None)

            return Packet(
                packet_id,
                packet_type,
                sender_id,
                receiver_id,
                assignment_id,
                data,
                conversation_id,
            )
        except Exception as e:
            print_and_log(
                logging.WARN,
                'Could not create a valid packet out of the dictionary'
                'provided: {}, error: {}'.format(packet, repr(e)),
            )
            return None

    def as_dict(self):
        """
        Convert a packet into a form that can be pushed over a socket.
        """
        return {
            'id': self.id,
            'type': self.type,
            'sender_id': self.sender_id,
            'receiver_id': self.receiver_id,
            'assignment_id': self.assignment_id,
            'conversation_id': self.conversation_id,
            'data': self.data,
        }

    def get_sender_connection_id(self):
        """
        Get the connection_id that this packet came from.
        """
        return '{}_{}'.format(self.sender_id, self.assignment_id)

    def get_receiver_connection_id(self):
        """
        Get the connection_id that this is going to.
        """
        return '{}_{}'.format(self.receiver_id, self.assignment_id)

    def new_copy(self):
        """
        Return a new packet that is a copy of this packet with a new id and with a fresh
        status.
        """
        packet = Packet.from_dict(self.as_dict())
        packet.id = shared_utils.generate_event_id(self.receiver_id)
        return packet

    def __repr__(self):
        return 'Packet <{}>'.format(self.as_dict())

    def swap_sender(self):
        """
        Swaps the sender_id and receiver_id.
        """
        self.sender_id, self.receiver_id = self.receiver_id, self.sender_id
        return self

    def set_type(self, new_type):
        """
        Updates the message type.
        """
        self.type = new_type
        return self

    def set_data(self, new_data):
        """
        Updates the message data.
        """
        self.data = new_data
        return self


class SocketManager:
    """
    SocketManager is a wrapper around websocket to conform to the API allowing the
    remote state to sync up with our local state.
    """

    # Default pings without pong before socket considered dead
    DEF_MISSED_PONGS: int = 20
    PING_RATE: float = 4
    DEF_DEAD_TIME: float = 30

    def __init__(
        self,
        server_url,
        port,
        alive_callback,
        message_callback,
        socket_dead_callback,
        task_group_id,
        socket_dead_timeout=None,
        server_death_callback=None,
    ):
        """
        server_url:           url at which the server is to be run
        port:                 port for the socket to operate on
        alive_callback:       function to be called on alive Packets, defined
                               alive_callback(self, pkt)
        message_callback:     function to be called on message Packets, defined
                               message_callback(self, pkt)
        socket_dead_callback: function to be called when a socket dies, should
                              return false if the socket_manager should ignore
                              the death and treat the socket as alive defined
                               on_socket_dead(self, worker_id, assignment_id)
        socket_dead_timeout:  time to wait between pings before dying
        """
        self.server_url = server_url
        self.port = port
        self.alive_callback = alive_callback
        self.message_callback = message_callback
        self.socket_dead_callback = socket_dead_callback
        self.server_death_callback = server_death_callback
        if socket_dead_timeout is not None:
            self.missed_pongs = 1 + socket_dead_timeout / self.PING_RATE
        else:
            self.missed_pongs = self.DEF_MISSED_PONGS
        self.task_group_id = task_group_id

        self.ws = None
        self.keep_running = True

        # initialize the state
        self.listen_thread = None
        self.send_thread = None
        self.sending_queue = PriorityQueue()
        self.open_channels = set()
        self.last_sent_ping_time = 0  # time of last ping send
        self.pings_without_pong = 0
        self.processed_packets = set()
        self.packet_map = {}
        self.alive = False
        self.is_shutdown = False
        self.send_lock = threading.Condition()
        self.packet_map_lock = threading.Condition()
        self.worker_assign_ids = {}  # mapping from connection id to pair

        # setup the socket
        self._setup_socket()

    def get_my_sender_id(self):
        """
        Gives the name that this socket manager should use for its world.
        """
        return '[World_{}]'.format(self.task_group_id)

    def _safe_send(self, data, force=False):
        if not self.alive and not force:
            # Try to wait a half second to send a packet
            timeout = 0.5
            while timeout > 0 and not self.alive:
                time.sleep(0.1)
                timeout -= 0.1
            if not self.alive:
                # don't try to send a packet if we're still dead
                return False
        try:
            with self.send_lock:
                self.ws.send(data)
        except websocket.WebSocketConnectionClosedException:
            # The channel died mid-send, wait for it to come back up
            return False
        except BrokenPipeError:  # noqa F821 we don't support p2
            # The channel died mid-send, wait for it to come back up
            return False
        except AttributeError:
            # _ensure_closed was called in parallel, self.ws = None
            return False
        except Exception as e:
            shared_utils.print_and_log(
                logging.WARN,
                'Unexpected socket error occured: {}'.format(repr(e)),
                should_print=True,
            )
            return False
        return True

    def _ensure_closed(self):
        self.alive = False
        if self.ws is None:
            return
        try:
            self.ws.close()
        except websocket.WebSocketConnectionClosedException:
            pass
        self.ws = None

    def _send_world_alive(self):
        """
        Registers world with the passthrough server.
        """
        self._safe_send(
            json.dumps(
                {
                    'type': data_model.AGENT_ALIVE,
                    'content': {
                        'id': 'WORLD_ALIVE',
                        'sender_id': self.get_my_sender_id(),
                    },
                }
            ),
            force=True,
        )

    def _try_send_world_ping(self):
        if time.time() - self.last_sent_ping_time > self.PING_RATE:
            self._safe_send(
                json.dumps(
                    {
                        'type': data_model.WORLD_PING,
                        'content': {
                            'id': 'WORLD_PING',
                            'sender_id': self.get_my_sender_id(),
                        },
                    }
                ),
                force=True,
            )
            self.last_sent_ping_time = time.time()

    def _send_packet(self, packet, send_time):
        """
        Sends a packet, blocks if the packet is blocking.
        """
        # Send the packet
        pkt = packet.as_dict()
        if pkt['data'] is None or packet.status == Packet.STATUS_SENT:
            return  # This packet was _just_ sent.
        shared_utils.print_and_log(logging.DEBUG, 'Send packet: {}'.format(packet))

        result = self._safe_send(json.dumps({'type': pkt['type'], 'content': pkt}))
        if not result:
            # The channel died mid-send, wait for it to come back up
            self.sending_queue.put((send_time, packet))
            return

        packet.status = Packet.STATUS_SENT
        if packet.ack_func is not None:
            packet.ack_func(packet)

    def _spawn_reaper_thread(self):
        def _reaper_thread(*args):
            start_time = time.time()
            wait_time = self.DEF_MISSED_PONGS * self.PING_RATE
            while time.time() - start_time < wait_time:
                if self.is_shutdown:
                    return
                if self.alive:
                    return
                time.sleep(0.3)
            if self.server_death_callback is not None:
                shared_utils.print_and_log(
                    logging.WARN,
                    'Server has disconnected and could not reconnect. '
                    'Assuming the worst and calling the death callback. '
                    '(Usually shutdown)',
                    should_print=True,
                )
                self.server_death_callback()

        reaper_thread = threading.Thread(
            target=_reaper_thread, name='socket-reaper-{}'.format(self.task_group_id)
        )
        reaper_thread.daemon = True
        reaper_thread.start()

    def _setup_socket(self):
        """
        Create socket handlers and registers the socket.
        """

        def on_socket_open(*args):
            shared_utils.print_and_log(logging.DEBUG, 'Socket open: {}'.format(args))
            self._send_world_alive()

        def on_error(ws, error):
            try:
                if error.errno == errno.ECONNREFUSED:
                    self.use_socket = False
                    self._ensure_closed()
                    raise Exception("Socket refused connection, cancelling")
                else:
                    shared_utils.print_and_log(
                        logging.WARN, 'Socket logged error: {}'.format(error)
                    )
                    self._ensure_closed()
            except Exception:
                if type(error) is websocket.WebSocketConnectionClosedException:
                    return  # Connection closed is noop
                shared_utils.print_and_log(
                    logging.WARN,
                    'Socket logged error: {} Restarting'.format(repr(error)),
                )
                self._ensure_closed()

        def on_disconnect(*args):
            """
            Disconnect event is a no-op for us, as the server reconnects automatically
            on a retry.

            Just in case the server is actually dead we set up a thread to reap the
            whole task.
            """
            shared_utils.print_and_log(
                logging.INFO, 'World server disconnected: {}'.format(args)
            )
            self._ensure_closed()
            if not self.is_shutdown:
                self._spawn_reaper_thread()

        def on_message(*args):
            """
            Incoming message handler for SERVER_PONG, MESSAGE_BATCH, AGENT_DISCONNECT,
            SNS_MESSAGE, SUBMIT_MESSAGE, AGENT_ALIVE.
            """
            packet_dict = json.loads(args[1])
            if packet_dict['type'] == 'conn_success':  # TODO make socket func
                self.alive = True
                return
            # The packet inherits the socket function type
            packet_dict['content']['type'] = packet_dict['type']
            packet = Packet.from_dict(packet_dict['content'])
            if packet is None:
                return
            packet_id = packet.id
            packet_type = packet.type
            if packet_id in self.processed_packets:
                return  # no need to handle already-processed packets

            # Note to self that this packet has already been processed,
            # and shouldn't be processed again in the future
            self.processed_packets.add(packet_id)
            if packet_type == data_model.SERVER_PONG:
                # Incoming pong means our ping was returned
                self.pings_without_pong = 0
            elif packet_type == data_model.AGENT_ALIVE:
                # agent is connecting for the first time
                self.alive_callback(packet)
                self.processed_packets.add(packet_id)
            elif packet_type == data_model.MESSAGE_BATCH:
                # Any number of agents are included in this message batch,
                # so process each individually
                batched_packets = packet.data['messages']
                for batched_packet_dict in batched_packets:
                    batched_packet_dict['type'] = data_model.AGENT_MESSAGE
                    batched_packet = Packet.from_dict(batched_packet_dict)
                    self.message_callback(batched_packet)
            elif packet_type == data_model.AGENT_DISCONNECT:
                # Server detected an agent disconnect, extract and remove
                disconnected_id = packet.data['connection_id']
                worker_id, assign_id = self.worker_assign_ids[disconnected_id]
                self.socket_dead_callback(worker_id, assign_id)
            elif packet_type == data_model.SNS_MESSAGE:
                # Treated as a regular message
                self.message_callback(packet)
            elif packet_type == data_model.SUBMIT_MESSAGE:
                # Treated as a regular message
                self.message_callback(packet)

        def run_socket(*args):
            url_base_name = self.server_url.split('https://')[1]
            protocol = "wss"
            if url_base_name in ['localhost', '127.0.0.1']:
                protocol = "ws"
            while self.keep_running:
                try:
                    sock_addr = "{}://{}:{}/".format(protocol, url_base_name, self.port)
                    self.ws = websocket.WebSocketApp(
                        sock_addr,
                        on_message=on_message,
                        on_error=on_error,
                        on_close=on_disconnect,
                    )
                    self.ws.on_open = on_socket_open
                    self.ws.run_forever(ping_interval=8 * self.PING_RATE)
                    self._ensure_closed()
                except Exception as e:
                    shared_utils.print_and_log(
                        logging.WARN,
                        'Socket error {}, attempting restart'.format(repr(e)),
                    )
                time.sleep(0.2)

        # Start listening thread
        self.listen_thread = threading.Thread(
            target=run_socket, name='Main-Socket-Recv-Thread'
        )
        self.listen_thread.daemon = True
        self.listen_thread.start()
        time.sleep(1.2)
        start_time = time.time()
        while not self.alive:
            if time.time() - start_time > self.DEF_DEAD_TIME:
                self.server_death_callback()
                raise ConnectionRefusedError(  # noqa F821 we only support py3
                    'Was not able to establish a connection with the server, '
                    'please try to run again. If that fails,'
                    'please ensure that your local device has the correct SSL '
                    'certs installed.'
                )
            try:
                self._send_world_alive()
            except Exception:
                pass
            time.sleep(self.PING_RATE / 2)

        # Start sending thread
        self.send_thread = threading.Thread(
            target=self.channel_thread, name='Main-Socket-Send-Thread'
        )
        self.send_thread.daemon = True
        self.send_thread.start()

    def channel_thread(self):
        """
        Handler thread for monitoring all channels.
        """
        # while the thread is still alive
        while not self.is_shutdown:
            if self.ws is None:
                # Waiting for websocket to come back alive
                time.sleep(shared_utils.THREAD_SHORT_SLEEP)
                continue
            self._try_send_world_ping()
            try:
                # Get first item in the queue, check if can send it yet
                item = self.sending_queue.get(block=False)
                t = item[0]
                if time.time() < t:
                    # Put the item back into the queue,
                    # it's not time to pop yet
                    self.sending_queue.put(item)
                else:
                    # Try to send the packet
                    packet = item[1]
                    if not packet:
                        # This packet was deleted out from under us
                        continue
                    if packet.status is not Packet.STATUS_SENT:
                        # either need to send initial packet
                        # or resend after a failed send
                        self._send_packet(packet, t)
            except Empty:
                time.sleep(shared_utils.THREAD_SHORT_SLEEP)
            except Exception as e:
                shared_utils.print_and_log(
                    logging.WARN,
                    'Unexpected error occurred in socket handling thread: '
                    '{}'.format(repr(e)),
                    should_print=True,
                )

    # Inividual channel accessors are useful for testing
    def open_channel(self, worker_id, assignment_id):
        """
        Opens a channel for a worker on a given assignment, doesn't re-open if the
        channel is already open.
        """
        connection_id = '{}_{}'.format(worker_id, assignment_id)
        self.open_channels.add(connection_id)
        self.worker_assign_ids[connection_id] = (worker_id, assignment_id)

    def close_channel(self, connection_id):
        """
        Closes a channel by connection_id.
        """
        shared_utils.print_and_log(
            logging.DEBUG, 'Closing channel {}'.format(connection_id)
        )
        if connection_id in self.open_channels:
            self.open_channels.remove(connection_id)
            with self.packet_map_lock:
                packet_ids = list(self.packet_map.keys())
                # Clear packets associated with this sender
                for packet_id in packet_ids:
                    packet = self.packet_map[packet_id]
                    packet_conn_id = packet.get_receiver_connection_id()
                    if connection_id == packet_conn_id:
                        del self.packet_map[packet_id]

    def close_all_channels(self):
        """
        Closes all channels by clearing the list of channels.
        """
        shared_utils.print_and_log(logging.DEBUG, 'Closing all channels')
        connection_ids = list(self.open_channels)
        for connection_id in connection_ids:
            self.close_channel(connection_id)

    def socket_is_open(self, connection_id):
        return connection_id in self.open_channels

    def queue_packet(self, packet):
        """
        Queues sending a packet to its intended owner.
        """
        shared_utils.print_and_log(
            logging.DEBUG, 'Put packet ({}) in queue'.format(packet.id)
        )
        # Get the current time to put packet into the priority queue
        with self.packet_map_lock:
            self.packet_map[packet.id] = packet
        item = (time.time(), packet)
        self.sending_queue.put(item)
        return True

    def get_status(self, packet_id):
        """
        Returns the status of a particular packet by id.
        """
        with self.packet_map_lock:
            if packet_id not in self.packet_map:
                return Packet.STATUS_NONE
            return self.packet_map[packet_id].status

    def shutdown(self):
        """
        marks the socket manager as closing, shuts down all channels.
        """
        self.is_shutdown = True
        self.close_all_channels()
        self.keep_running = False
        self._ensure_closed()



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from parlai.core.params import ParlaiParser
from parlai.core.agents import create_agent
from parlai.mturk.core.legacy_2018.mturk_manager import MTurkManager
import parlai.mturk.core.legacy_2018.mturk_utils as mturk_utils

from worlds import ControllableDialogEval, PersonasGenerator, PersonaAssignWorld
from task_config import task_config
import model_configs as mcf

from threading import Lock
import gc
import datetime
import json
import os
import sys
import copy
import random
import pprint
from parlai.utils.logging import ParlaiLogger, INFO


# update this with models you want to run. these names correspond to variables
# in model_configs.py
SETTINGS_TO_RUN = """
baseline_model
greedy_model
inquisitive_model_ct_setting00
inquisitive_model_ct_setting01
inquisitive_model_ct_setting04
inquisitive_model_ct_setting07
inquisitive_model_ct_setting10
inquisitive_model_ct_setting10_better
interesting_nidf_model_bfw_setting_04
interesting_nidf_model_bfw_setting_06
interesting_nidf_model_bfw_setting_08
interesting_nidf_model_bfw_setting_minus_04
interesting_nidf_model_bfw_setting_minus_10
interesting_nidf_model_ct_setting0
interesting_nidf_model_ct_setting2
interesting_nidf_model_ct_setting4
interesting_nidf_model_ct_setting7
interesting_nidf_model_ct_setting9
repetition_model_setting05
repetition_model_setting12
repetition_model_setting35
repetition_model_setting35_settinginf
repetition_model_settinginf
responsiveness_model_bfw_setting_00
responsiveness_model_bfw_setting_05
responsiveness_model_bfw_setting_10
responsiveness_model_bfw_setting_13
responsiveness_model_bfw_setting_minus_10
""".strip().split()


def main():
    """
    This task consists of an MTurk agent evaluating a Controllable Dialog model.
    """
    start_time = datetime.datetime.today().strftime('%Y-%m-%d-%H-%M')
    argparser = ParlaiParser(False, add_model_args=True)
    argparser.add_parlai_data_path()
    argparser.add_mturk_args()
    argparser.add_argument(
        '--max-resp-time',
        default=240,
        type=int,
        help='time limit for entering a dialog message',
    )
    argparser.add_argument(
        '--max-choice-time',
        type=int,
        default=300,
        help='time limit for turker' 'choosing the topic',
    )
    argparser.add_argument(
        '--ag-shutdown-time',
        default=120,
        type=int,
        help='time limit for entering a dialog message',
    )
    argparser.add_argument(
        '--num-turns', default=6, type=int, help='number of turns of dialogue'
    )
    argparser.add_argument(
        '--human-eval',
        type='bool',
        default=False,
        help='human vs human eval, no models involved',
    )
    argparser.add_argument(
        '--auto-approve-delay',
        type=int,
        default=3600 * 24 * 2,
        help='how long to wait for auto approval',
    )
    argparser.add_argument(
        '--only-masters',
        type='bool',
        default=False,
        help='Set to true to use only master turks for ' 'this test eval',
    )
    argparser.add_argument(
        '--create-model-qualif',
        type='bool',
        default=True,
        help='Create model qualif so unique eval between' 'models.',
    )
    argparser.add_argument(
        '--limit-workers',
        type=int,
        default=len(SETTINGS_TO_RUN),
        help='max HITs a worker can complete',
    )
    argparser.add_argument(
        '--mturk-log',
        type=str,
        default=('data/mturklogs/controllable/{}.log'.format(start_time)),
    )
    argparser.add_argument(
        '--short-eval',
        type='bool',
        default=True,
        help='Only ask engagingness question and persona' 'question.',
    )
    # persona specific arguments
    argparser.add_argument(
        '--persona-type', type=str, default='self', choices=['self', 'other', 'none']
    )
    argparser.add_argument(
        '--persona-datatype',
        type=str,
        default='valid',
        choices=['train', 'test', 'valid'],
    )
    argparser.add_argument(
        '--max-persona-time', type=int, default=360, help='max time to view persona'
    )

    def get_logger(opt):
        fmt = '%(asctime)s: [ %(message)s ]'
        logfn = None
        if 'mturk_log' in opt:
            logfn = opt['mturk_log']
            if not os.path.isdir(os.path.dirname(logfn)):
                os.makedirs(os.path.dirname(logfn), exist_ok=True)
        logger = ParlaiLogger(
            name="mturk_controllable",
            console_level=INFO,
            file_level=INFO,
            console_format=fmt,
            file_format=fmt,
            filename=logfn,
        )
        logger.info('COMMAND: %s' % ' '.join(sys.argv))
        logger.info('-' * 100)
        logger.info('CONFIG:\n%s' % json.dumps(opt, indent=4, sort_keys=True))

        return logger

    start_opt = argparser.parse_args()

    task_config['task_description'] = task_config['task_description'].format(
        start_opt['reward']
    )

    # set options
    start_opt['limit_workers'] = len(SETTINGS_TO_RUN)
    start_opt['allowed_conversations'] = 1
    start_opt['max_hits_per_worker'] = start_opt['limit_workers']
    start_opt['task'] = os.path.basename(os.path.dirname(os.path.abspath(__file__)))

    start_opt.update(task_config)

    logger = get_logger(start_opt)

    model_share_params = {}
    worker_models_seen = {}
    model_opts = {}
    model_counts = {}

    lock = Lock()

    for setup in SETTINGS_TO_RUN:
        assert 'human' not in setup
        model_counts[setup] = 0
        agent_config = getattr(mcf, setup)
        combined_config = copy.deepcopy(start_opt)
        for k, v in agent_config.items():
            combined_config[k] = v
            combined_config['override'][k] = v
        folder_name = '{}-{}'.format(setup, start_time)
        combined_config['save_data_path'] = os.path.join(
            start_opt['datapath'], 'local_controllable_dialogue', folder_name
        )
        model_opts[setup] = combined_config
        bot = create_agent(combined_config, True)
        model_share_params[setup] = bot.share()

    if not start_opt.get('human_eval'):
        mturk_agent_ids = ['PERSON_1']
    else:
        mturk_agent_ids = ['PERSON_1', 'PERSON_2']

    mturk_manager = MTurkManager(opt=start_opt, mturk_agent_ids=mturk_agent_ids)

    personas_generator = PersonasGenerator(start_opt)

    directory_path = os.path.dirname(os.path.abspath(__file__))

    mturk_manager.setup_server(task_directory_path=directory_path)

    try:
        mturk_manager.start_new_run()
        agent_qualifications = []
        # assign qualifications
        if start_opt['create_model_qualif']:
            qual_name = 'ControlEvalRound2'
            qual_desc = (
                'Qualification to ensure workers complete only a certain'
                'number of these HITs'
            )
            qualification_id = mturk_utils.find_or_create_qualification(
                qual_name, qual_desc, False
            )
            print('Created qualification: ', qualification_id)
            start_opt['unique_qualif_id'] = qualification_id

        def run_onboard(worker):
            worker.personas_generator = personas_generator
            world = PersonaAssignWorld(start_opt, worker)
            world.parley()
            world.shutdown()

        def check_worker_eligibility(worker):
            worker_id = worker.worker_id
            lock.acquire()
            retval = len(worker_models_seen.get(worker_id, [])) < len(SETTINGS_TO_RUN)
            lock.release()
            return retval

        def assign_worker_roles(workers):
            for index, worker in enumerate(workers):
                worker.id = mturk_agent_ids[index % len(mturk_agent_ids)]

        mturk_manager.set_onboard_function(onboard_function=run_onboard)
        mturk_manager.ready_to_accept_workers()
        mturk_manager.create_hits(qualifications=agent_qualifications)

        def run_conversation(mturk_manager, opt, workers):
            conv_idx = mturk_manager.conversation_index

            # gotta find a bot this worker hasn't seen yet
            assert len(workers) == 1
            worker_id = workers[0].worker_id
            lock.acquire()
            if worker_id not in worker_models_seen:
                worker_models_seen[worker_id] = set()
            print("MODELCOUNTS:")
            print(pprint.pformat(model_counts))
            logger.info("MODELCOUNTS\n" + pprint.pformat(model_counts))
            model_options = [
                (model_counts[setup_name] + 10 * random.random(), setup_name)
                for setup_name in SETTINGS_TO_RUN
                if setup_name not in worker_models_seen[worker_id]
            ]
            if not model_options:
                lock.release()
                logger.error(
                    "Worker {} already finished all settings! Returning none".format(
                        worker_id
                    )
                )
                return None
            _, model_choice = min(model_options)

            worker_models_seen[worker_id].add(model_choice)
            model_counts[model_choice] += 1
            lock.release()

            world = ControllableDialogEval(
                opt=model_opts[model_choice],
                agents=workers,
                num_turns=start_opt['num_turns'],
                max_resp_time=start_opt['max_resp_time'],
                model_agent_opt=model_share_params[model_choice],
                world_tag='conversation t_{}'.format(conv_idx),
                agent_timeout_shutdown=opt['ag_shutdown_time'],
                model_config=model_choice,
            )
            world.reset_random()
            while not world.episode_done():
                world.parley()
            world.save_data()

            lock.acquire()
            if not world.convo_finished:
                model_counts[model_choice] -= 1
                worker_models_seen[worker_id].remove(model_choice)
            lock.release()

            world.shutdown()
            gc.collect()

        mturk_manager.start_task(
            eligibility_function=check_worker_eligibility,
            assign_role_function=assign_worker_roles,
            task_function=run_conversation,
        )

    except BaseException:
        raise
    finally:
        mturk_manager.expire_all_unassigned_hits()
        mturk_manager.shutdown()


if __name__ == '__main__':
    main()



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
from parlai.core.params import ParlaiParser
from parlai.mturk.tasks.qa_data_collection.worlds import (
    QADataCollectionOnboardWorld,
    QADataCollectionWorld,
)
from parlai.mturk.core.mturk_manager import MTurkManager
from parlai.mturk.tasks.qa_data_collection.task_config import task_config
import os
import importlib


def main():
    """
    Handles setting up and running a ParlAI-MTurk task by instantiating an MTurk manager
    and configuring it for the qa_data_collection task.
    """
    # Get relevant arguments
    argparser = ParlaiParser(False, False)
    argparser.add_parlai_data_path()
    argparser.add_mturk_args()
    opt = argparser.parse_args()

    # Set the task name to be the folder name
    opt['task'] = os.path.basename(os.path.dirname(os.path.abspath(__file__)))

    # append the contents of task_config.py to the configuration
    opt.update(task_config)

    # Initialize a SQuAD teacher agent, which we will get context from
    module_name = 'parlai.tasks.squad.agents'
    class_name = 'DefaultTeacher'
    my_module = importlib.import_module(module_name)
    task_class = getattr(my_module, class_name)
    task_opt = opt.copy()
    task_opt['datatype'] = 'train'
    task_opt['datapath'] = opt['datapath']

    # Select an agent_id that worker agents will be assigned in their world
    mturk_agent_id = 'Worker'

    # Instantiate an MTurkManager with the given options and a maximum number
    # of agents per world of 1 (based on the length of mturk_agent_ids)
    mturk_manager = MTurkManager(opt=opt, mturk_agent_ids=[mturk_agent_id], use_db=True)
    mturk_manager.setup_server()

    # Create an onboard_function, which will be run for workers who have
    # accepted your task and must be completed before they are put in the
    # queue for a task world.
    def run_onboard(worker):
        world = QADataCollectionOnboardWorld(opt=opt, mturk_agent=worker)
        while not world.episode_done():
            world.parley()
        world.shutdown()
        return world.prep_save_data([worker])

    # If we want to use the above onboard function, we can replace the below
    # with set_onboard_function(onboard_function=run_onboard)
    mturk_manager.set_onboard_function(onboard_function=None)

    try:
        # Initialize run information
        mturk_manager.start_new_run()

        # Set up the sockets and threads to recieve workers
        mturk_manager.ready_to_accept_workers()

        # Create the hits as specified by command line arguments
        mturk_manager.create_hits()

        # Check workers eligiblity acts as a filter, and should return
        # the list of all workers currently eligible to work on the task
        def check_workers_eligibility(workers):
            return workers

        eligibility_function = {'func': check_workers_eligibility, 'multiple': True}

        # Assign worker roles is used to determine what the role each worker
        # in the given worker list will play. Setting `id` to None will return
        # the worker to the pool rather than putting them in a given task,
        # which is useful for having tasks with different possible worker
        # counts.
        def assign_worker_roles(workers):
            workers[0].id = mturk_agent_id

        # Define the task function, which will be run with workers that are
        # as the main task.
        global run_conversation

        def run_conversation(mturk_manager, opt, workers):
            # create a task agent to ask the questions
            task = task_class(task_opt)
            # Create the task world
            world = QADataCollectionWorld(opt=opt, task=task, mturk_agent=workers[0])
            # run the world to completion
            while not world.episode_done():
                world.parley()

            # shutdown and review the work
            world.shutdown()
            world.review_work()

            # Return the contents for saving
            return world.prep_save_data(workers)

        # Begin the task, allowing mturk_manager to start running the task
        # world on any workers who connect
        mturk_manager.start_task(
            eligibility_function=eligibility_function,
            assign_role_function=assign_worker_roles,
            task_function=run_conversation,
        )
    except BaseException:
        raise
    finally:
        # Any hits that aren't claimed or completed have to be shut down. Must
        # keep the world running until that point.
        mturk_manager.expire_all_unassigned_hits()
        # Shutdown the manager and free all related resources
        mturk_manager.shutdown()


if __name__ == '__main__':
    main()



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
#
# Download and build the data if it does not exist.

from parlai.core.teachers import MultiTaskTeacher
import parlai.tasks.anli.agents as anli
import parlai.tasks.multinli.agents as multinli
import parlai.tasks.snli.agents as snli
import parlai.tasks.dialogue_nli.agents as dnli
from copy import deepcopy


class AnliTeacher(anli.DefaultTeacher):
    pass


class MultinliTeacher(multinli.DefaultTeacher):
    pass


class SnliTeacher(snli.DefaultTeacher):
    pass


class DnliTeacher(dnli.DefaultTeacher):
    pass


class NliTeacher(MultiTaskTeacher):
    @staticmethod
    def add_cmdline_args(parser):
        parser = parser.add_argument_group('NLI Teacher Args')
        parser.add_argument(
            '-dfm',
            '--dialog-format',
            type='bool',
            default=False,
            help="True if one would like to convert to a dialogue format without special tokens such as 'Premise'"
            " and 'Hypothesis' (default: False).",
        )
        parser.add_argument(
            '-bcl',
            '--binary-classes',
            type='bool',
            default=False,
            help="True if label candidates are (contradiction, not_contradiction), and (entailment, contradiction, "
            "neutral) otherwise (default: False).",
        )

    def __init__(self, opt, shared=None):
        nli_tasks = [
            'anli:r1',
            'anli:r2',
            'anli:r3',
            'multinli',
            'snli',
            'dialogue_nli',
        ]
        opt = deepcopy(opt)
        opt['task'] = ','.join(nli_tasks)
        super().__init__(opt, shared)


class DefaultTeacher(NliTeacher):
    pass



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
from parlai.mturk.core.worlds import MTurkTaskWorld
from parlai.core.worlds import validate
from joblib import Parallel, delayed
from parlai.tasks.dealnodeal.agents import NegotiationTeacher
from parlai.tasks.dealnodeal.agents import get_tag
from parlai.tasks.dealnodeal.agents import WELCOME_MESSAGE

import random


class MTurkDealNoDealDialogWorld(MTurkTaskWorld):
    """
    World where two agents have a dialogue to negotiate a deal.
    """

    def __init__(self, opt, agents=None, shared=None):
        # Add passed in agents directly.
        if agents is not None:
            random.shuffle(agents)
        self.agents = agents
        self.acts = [None] * len(agents)
        self.episodeDone = False
        self.task = NegotiationTeacher(opt=opt)
        self.first_turn = True
        self.choices = dict()
        self.selection = False
        self.turns = 0
        self.num_negotiations = 0

    def parley(self):
        """
        Alternate taking turns, until both agents have made a choice (indicated by a
        turn starting with <selection>)
        """
        if self.first_turn:
            # Use NegotiationTeacher to load data for us
            data = (
                self.task.episodes[self.num_negotiations % len(self.task.episodes)]
                .strip()
                .split()
            )
            self.num_negotiations += 1

            for agent, tag in zip(self.agents, ['input', 'partner_input']):
                (book_cnt, book_val, hat_cnt, hat_val, ball_cnt, ball_val) = get_tag(
                    data, tag
                )
                action = {}
                action['text'] = WELCOME_MESSAGE.format(
                    book_cnt=book_cnt,
                    book_val=book_val,
                    hat_cnt=hat_cnt,
                    hat_val=hat_val,
                    ball_cnt=ball_cnt,
                    ball_val=ball_val,
                )

                action['items'] = {
                    "book_cnt": book_cnt,
                    "book_val": book_val,
                    "hat_cnt": hat_cnt,
                    "hat_val": hat_val,
                    "ball_cnt": ball_cnt,
                    "ball_val": ball_val,
                }

                agent.observe(validate(action))
            self.first_turn = False
        else:
            self.turns += 1

            for _index, agent in enumerate(self.agents):
                if agent in self.choices:
                    # This agent has already made a choice
                    continue

                try:
                    act = agent.act(timeout=None)
                except TypeError:
                    act = agent.act()  # not MTurkAgent

                for other_agent in self.agents:
                    if other_agent != agent:
                        other_agent.observe(validate(act))

                if act["text"].startswith("<selection>") and self.turns > 1:
                    # Making a choice
                    self.choices[agent] = act["text"]
                    self.selection = True
                    if len(self.choices) == len(self.agents):
                        self.first_turn = True
                        self.episodeDone = True
                elif act['episode_done']:
                    # Action is not selection but episode ended due to
                    # disconnection or timeout or returned hit
                    self.episodeDone = True

    def episode_done(self):
        return self.episodeDone

    def shutdown(self):
        """
        Shutdown all mturk agents in parallel, otherwise if one mturk agent is
        disconnected then it could prevent other mturk agents from completing.
        """
        global shutdown_agent

        def shutdown_agent(agent):
            try:
                agent.shutdown(timeout=None)
            except Exception:
                agent.shutdown()  # not MTurkAgent

        Parallel(n_jobs=len(self.agents), backend='threading')(
            delayed(shutdown_agent)(agent) for agent in self.agents
        )
